Package: nprobust
Type: Package
Title: Kernel Density and Local Polynomial Regression Methods
Version: 1.0.0
Date: 2026-05-17
Authors@R: c(person(given = c("Sebastian"),
                      family = "Calonico",
                      role = c("aut", "cre"),
                      email = "scalonico@ucdavis.edu"),
               person(given = c("Matias", "D."),
                      family = "Cattaneo",
                      role = "aut",
                      email = "matias.d.cattaneo@gmail.com"),
               person(given = c("Max", "H."),
                      family = "Farrell",
                      role = "aut",
                      email = "mhfarrell@gmail.com"))
Author: Sebastian Calonico [aut, cre],
  Matias D. Cattaneo [aut],
  Max H. Farrell [aut]
Maintainer: Sebastian Calonico <scalonico@ucdavis.edu>
Description: Estimation, inference, bandwidth selection, and graphical procedures for kernel density and local polynomial regression methods, including robust bias-corrected confidence intervals as described in Calonico, Cattaneo and Farrell (2018, <doi:10.1080/01621459.2017.1285776>). The package includes 'lprobust()' for local polynomial point estimation and robust bias-corrected inference, 'lpbwselect()' for local polynomial bandwidth selection, 'kdrobust()' for kernel density point estimation and robust bias-corrected inference, 'kdbwselect()' for kernel density bandwidth selection, and 'nprobust.plot()' for plotting results. The main methodological and numerical features are described in Calonico, Cattaneo and Farrell (2019, <doi:10.18637/jss.v091.i08>).
Encoding: UTF-8
Depends: R (>= 3.6.0)
License: GPL-3
URL: https://github.com/nppackages/nprobust
BugReports: https://github.com/nppackages/nprobust/issues
Imports: ggplot2
Suggests: testthat (>= 3.0.0), broom, sandwich
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-18 21:54:24 UTC; ncalonic
Repository: CRAN
Date/Publication: 2026-05-19 05:10:09 UTC
Built: R 4.5.2; ; 2026-05-19 10:29:27 UTC; unix

S3method(print,lprobust)
S3method(summary,lprobust)
S3method(print,summary.lprobust)
S3method(plot,lprobust)
S3method(print,lpbwselect)
S3method(summary,lpbwselect)
S3method(print,summary.lpbwselect)
S3method(print,kdrobust)
S3method(summary,kdrobust)
S3method(print,summary.kdrobust)
S3method(plot,kdrobust)
S3method(print,kdbwselect)
S3method(summary,kdbwselect)
S3method(print,summary.kdbwselect)
S3method(ggplot2::autoplot,lprobust)
S3method(ggplot2::autoplot,kdrobust)
S3method(broom::tidy,lprobust)
S3method(broom::glance,lprobust)
S3method(broom::tidy,kdrobust)
S3method(broom::glance,kdrobust)
S3method(broom::tidy,lpbwselect)
S3method(broom::glance,lpbwselect)
S3method(broom::tidy,kdbwselect)
S3method(broom::glance,kdbwselect)
export(lpbwselect)
export(lprobust)
export(kdbwselect)
export(kdrobust)
export(nprobust.plot)
import(ggplot2)
importFrom("stats","integrate","complete.cases","deriv","dnorm","kernel","lm","optimize","pnorm","poly","qnorm","quantile","sd","sigma")
importFrom("graphics","lines","plot","polygon")

# Kernel Density and Local Polynomial Regression Methods

The package `nprobust` implements estimation, inference, bandwidth selection,
and graphical procedures for kernel density and local polynomial regression
methods, including robust bias-corrected confidence intervals.

- `lprobust()`: local polynomial point estimation and robust bias-corrected inference.
- `lpbwselect()`: data-driven bandwidth selection for local polynomial regression.
- `kdrobust()`: kernel density point estimation and robust bias-corrected inference.
- `kdbwselect()`: data-driven bandwidth selection for kernel density estimation.
- `nprobust.plot()`: graphical presentation of `lprobust()` and `kdrobust()` results.

See references for methodological and practical details.

Website: [https://nppackages.github.io/](https://nppackages.github.io/).

Source code: [https://github.com/nppackages/nprobust](https://github.com/nppackages/nprobust).

## Authors

Sebastian Calonico (<scalonico@ucdavis.edu>)

Matias D. Cattaneo (<matias.d.cattaneo@gmail.com>)

Max H. Farrell (<mhfarrell@gmail.com>)

## Installation

To install/update use R:

```r
install.packages("nprobust")
```

## Usage

```r
library(nprobust)

# Cholesterol trial data used by the Python and Stata examples.
data <- read.csv("../nprobust_data.csv")
control <- data$t == 0

# Local polynomial regression with robust bias-corrected confidence intervals.
result <- lprobust(data$cholf[control], data$chol1[control])
summary(result)

# Data-driven bandwidth selection.
bw <- lpbwselect(data$cholf[control], data$chol1[control],
                 bwselect = "mse-dpi", neval = 7)
summary(bw)

# Kernel density estimation.
density <- kdrobust(data$chol1[control], neval = 30)
summary(density)

# Kernel density bandwidth selection.
summary(kdbwselect(data$chol1[control], bwselect = "imse-dpi"))

# Plot a local polynomial fit.
nprobust.plot(result, xlabel = "chol1", ylabel = "cholf")
```

- Replication: [nprobust illustration](https://github.com/nppackages/nprobust/blob/main/R/nprobust_illustration.R), [nprobust data](https://github.com/nppackages/nprobust/blob/main/R/nprobust_data.csv).

## Dependencies

- ggplot2

## References

For overviews and introductions, see [nppackages website](https://nppackages.github.io).

### Software and Implementation

- Calonico, Cattaneo and Farrell (2019): [nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference](https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf).<br>
_Journal of Statistical Software_ 91(8): 1-33.

### Technical and Methodological

- Calonico, Cattaneo and Farrell (2018): [On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference](https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf).<br>
_Journal of the American Statistical Association_ 113(522): 767-779.

- Calonico, Cattaneo and Farrell (2022): [Coverage Error Optimal Confidence Intervals for Local Polynomial Regression](https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2022_Bernoulli.pdf).<br>
_Bernoulli_ 28(4): 2998-3022.

## S3 plot() and autoplot() wrappers
plot.lprobust <- function(x, ...) nprobust.plot(x, ...)
plot.kdrobust <- function(x, ...) nprobust.plot(x, ...)

autoplot.lprobust <- function(object, ...) nprobust.plot(object, ...)
autoplot.kdrobust <- function(object, ...) nprobust.plot(object, ...)

nprobust.plot <- function(..., alpha=NULL, type=NULL, CItype=NULL,
                          title="", xlabel="", ylabel="",
                          lty=NULL, lwd=NULL, lcol=NULL, pty=NULL, pwd=NULL, pcol=NULL,
                          CIshade=NULL, CIcol=NULL, legendTitle=NULL, legendGroups=NULL) {

  ########################################
  # check how many series are passed in 
  ########################################

  x <- list(...)
  nfig <- length(x)
  if (nfig == 0) stop("Nothing to plot.", call. = FALSE)

  # Type guard: only lprobust/kdrobust outputs have an Estimate matrix
  bad <- !vapply(x, inherits, logical(1), what = c("lprobust", "kdrobust"))
  if (any(bad))
    stop("nprobust.plot() inputs must be 'lprobust' or 'kdrobust' results; ",
         "got ", paste(unique(unlist(lapply(x[bad], class))), collapse=", "), ".",
         call. = FALSE)

  ########################################
  # error handling
  ########################################
  # alpha
  if (length(alpha) == 0) {
    alpha <- rep(0.05, nfig)
  } else if (!is.numeric(alpha) || any(!is.finite(alpha)) || !all(alpha > 0 & alpha < 1)) {
    stop("alpha must be numeric with all entries in (0, 1).", call. = FALSE)
  } else {
    alpha <- rep(alpha, length.out=nfig)
  }

  # plot type
  if (length(type) == 0) {
    type <- rep("line", nfig)
  } else {
    if (!all(type %in% c("line", "points", "both"))) {
      stop("type must be one of \"line\", \"points\", \"both\".", call. = FALSE)
    }
    type <- rep(type, length.out=nfig)
  }

  # CI type
  if (length(CItype) == 0) {
    CItype <- rep("region", nfig)
  } else {
    if (!all(CItype %in% c("region", "line", "ebar", "all", "none"))) {
      stop("CItype must be one of \"region\", \"line\", \"ebar\", \"all\", \"none\".", call. = FALSE)
    }
    CItype <- rep(CItype, length.out=nfig)
  }

  # line style, line width, line color
  if (length(lty) == 0) {
    lty <- rep(1, nfig)
  } else {
    lty <- rep(lty, length.out=nfig)
  }
  if (length(lwd) == 0) {
    lwd <- rep(0.5, nfig)
  } else {
    lwd <- rep(lwd, length.out=nfig)
  }
  if (length(lcol) == 0) {
    lcol <- 1:nfig
  } else {
    lcol <- rep(lcol, length.out=nfig)
  }

  # point style, point width, point color
  if (length(pty) == 0) {
    pty <- rep(1, nfig)
  } else {
    pty <- rep(pty, length.out=nfig)
  }
  if (length(pwd) == 0) {
    pwd <- rep(1, nfig)
  } else {
    pwd <- rep(pwd, length.out=nfig)
  }
  if (length(pcol) == 0) {
    pcol <- lcol
  } else {
    pcol <- rep(pcol, length.out=nfig)
  }

  # CI shade, CI color
  if (length(CIshade) == 0) {
    CIshade <- rep(0.2, nfig)
  } else {
    CIshade <- rep(CIshade, length.out=nfig)
  }
  if (length(CIcol) == 0) {
    CIcol <- lcol
  } else {
    CIcol <- rep(CIcol, length.out=nfig)
  }

  # legend
  # Changes made by Xinwei
  if (length(legendTitle) == 0) {
    legendTitle <- ""
  } else {
    legendTitle <- legendTitle[1]
  }
  if (length(legendGroups) > 0) {
    legendGroups <- rep(legendGroups, length.out=nfig)
    legend_default <- FALSE
  } else {
    legend_default <- TRUE
  }

  ########################################
  # initializing plot
  ########################################
  temp_plot <- ggplot() + theme_bw() #+ theme(legend.position="none")

  CI_l <- CI_r <- tau.us <- eval <- Sname <- NULL

  ########################################
  # looping over input models
  ########################################
  ### all colors
  col_all <- lty_all <- pty_all <- c()
  for (i in 1:nfig) {
    data_x <- data.frame(x[[i]]$Estimate[, c("eval", "tau.us", "tau.bc", "se.us", "se.rb")])
    z_val <- qnorm(1 - alpha[i]/2)
    data_x$CI_l <- data_x$tau.bc - z_val * data_x$se.rb
    data_x$CI_r <- data_x$tau.bc + z_val * data_x$se.rb

    # changes made by Xinwei
    if (legend_default) {
      data_x$Sname <- paste("Series", i, sep=" ")
      legendGroups <- c(legendGroups, data_x$Sname)
    } else {
      data_x$Sname <- legendGroups[i]
    }

    ########################################
    # add CI regions to the plot
    if (CItype[i]%in%c("region", "all"))
      temp_plot <- temp_plot + geom_ribbon(data=data_x, aes(x=eval, ymin=CI_l, ymax=CI_r), alpha=CIshade[i], fill=CIcol[i])

    ########################################
    # add CI lines to the plot
    if (CItype[i]%in%c("line", "all"))
      temp_plot <- temp_plot + geom_line(data=data_x, aes(x=eval, y=CI_l), linetype=2, alpha=CIshade[i], col=CIcol[i]) +
      geom_line(data=data_x, aes(x=eval, y=CI_r), linetype=2, alpha=CIshade[i], col=CIcol[i])

    ########################################
    # add error bars to the plot
    if (CItype[i]%in%c("ebar", "all"))
      temp_plot <- temp_plot + geom_errorbar(data=data_x, aes(x=eval, ymin=CI_l, ymax=CI_r), alpha=CIshade[i], col=CIcol[i], linetype=1)

    ########################################
    # add lines to the plot
    if (type[i]%in%c("line", "both")) {
      temp_plot <- temp_plot + geom_line(data=data_x, aes(x=eval, y=tau.us, colour=Sname, linetype=Sname), linewidth=lwd[i])
    }

    ########################################
    # add points to the plot
    if (type[i]%in%c("points", "both")) {
      temp_plot <- temp_plot + geom_point(data=data_x, aes(x=eval, y=tau.us, colour=Sname, shape=Sname), size=pwd[i])
    }

    if (type[i] == "line") {
      col_all <- c(col_all, lcol[i])
      lty_all <- c(lty_all, lty[i])
      pty_all <- c(pty_all, NA)
    } else if (type[i] == "both") {
      col_all <- c(col_all, lcol[i])
      lty_all <- c(lty_all, lty[i])
      pty_all <- c(pty_all, pty[i])
    } else {
      col_all <- c(col_all, pcol[i])
      lty_all <- c(lty_all, NA)
      pty_all <- c(pty_all, pty[i])
    }
  }

  ########################################
  # change color, line type and point shape back, and customize legend
  ########################################
  
  index <- sort.int(legendGroups, index.return=TRUE)$ix
  temp_plot <- temp_plot + scale_color_manual(values = col_all[index]) +
    scale_linetype_manual(values = lty_all[index]) +
    scale_shape_manual(values = pty_all[index]) +
    guides(colour=guide_legend(title=legendTitle)) +
    guides(linetype=guide_legend(title=legendTitle)) +
    guides(shape=guide_legend(title=legendTitle))

  ########################################
  # add title, x and y labs
  ########################################
  temp_plot <- temp_plot + labs(x=xlabel, y=ylabel) + ggtitle(title)
 
  ########################################
  # return the plot
  ########################################
  return (temp_plot)
}

## S3 plot methods should dispatch to nprobust.plot and return ggplot objects.

test_that("plot.lprobust returns a ggplot", {
  fx <- make_fixture(n = 300)
  m  <- lprobust(fx$y, fx$x, neval = 10)
  g  <- plot(m)
  expect_s3_class(g, "ggplot")
})

test_that("plot.kdrobust returns a ggplot", {
  fx <- make_fixture(n = 300)
  kd <- kdrobust(fx$x, neval = 10)
  g  <- plot(kd)
  expect_s3_class(g, "ggplot")
})

test_that("nprobust.plot accepts multiple series", {
  fx <- make_fixture(n = 400)
  m0 <- lprobust(fx$y[fx$x <= 0], fx$x[fx$x <= 0], eval = seq(-0.8, 0, length.out = 10))
  m1 <- lprobust(fx$y[fx$x >  0], fx$x[fx$x >  0], eval = seq(0, 0.8, length.out = 10))
  g  <- nprobust.plot(m0, m1, legendGroups = c("Left", "Right"))
  expect_s3_class(g, "ggplot")
})

## Shared kernel helper + fixture dataset
w_fun <- function(u, kernel) {
  switch(kernel,
         "uni" = 0.5 * (abs(u) <= 1),
         "tri" = (1 - abs(u)) * (abs(u) <= 1),
         "epa" = 0.75 * (1 - u^2) * (abs(u) <= 1))
}

make_fixture <- function(n = 1000, seed = 7) {
  set.seed(seed)
  x <- sort(runif(n, -1, 1))
  y <- 1 + 2 * x + 3 * x^2 - 0.5 * x^3 + rnorm(n, sd = 0.3)
  list(x = x, y = y, n = n)
}

## print/summary contracts:
##   - print(x) returns x invisibly
##   - summary(x) returns an S3 object (class "summary.<class>")
##   - print(summary(x)) returns its argument invisibly

test_that("print methods return their input invisibly", {
  fx <- make_fixture(n = 300)
  m  <- lprobust(fx$y, fx$x, eval = c(0, 0.5))
  bw <- lpbwselect(fx$y, fx$x, eval = c(0, 0.5))
  kd <- kdrobust(fx$x, eval = c(0, 0.5))
  kb <- kdbwselect(fx$x, eval = c(0, 0.5))

  silence <- file(tempfile(), open = "wt")
  sink(silence)
  on.exit({ sink(); close(silence) }, add = TRUE)

  for (obj in list(m, bw, kd, kb)) {
    res <- withVisible(print(obj))
    expect_false(res$visible)
    expect_identical(res$value, obj)
  }
})

test_that("summary methods return an S3 summary object", {
  fx <- make_fixture(n = 300)
  m  <- lprobust(fx$y, fx$x, eval = c(0, 0.5))
  bw <- lpbwselect(fx$y, fx$x, eval = c(0, 0.5))
  kd <- kdrobust(fx$x, eval = c(0, 0.5))
  kb <- kdbwselect(fx$x, eval = c(0, 0.5))

  capture.output({
    s.m  <- summary(m)
    s.bw <- summary(bw)
    s.kd <- summary(kd)
    s.kb <- summary(kb)
  })

  expect_s3_class(s.m,  "summary.lprobust")
  expect_s3_class(s.bw, "summary.lpbwselect")
  expect_s3_class(s.kd, "summary.kdrobust")
  expect_s3_class(s.kb, "summary.kdbwselect")

  ## CI bounds inside the summary match tau.bc +/- z*se.rb
  z <- qnorm(1 - 0.05 / 2)
  expect_equal(s.m$CI_l,  s.m$Estimate[, "tau.bc"]  - z * s.m$Estimate[, "se.rb"])
  expect_equal(s.m$CI_r,  s.m$Estimate[, "tau.bc"]  + z * s.m$Estimate[, "se.rb"])
  expect_equal(s.kd$CI_l, s.kd$Estimate[, "tau.bc"] - z * s.kd$Estimate[, "se.rb"])
})

test_that("summary respects alpha argument", {
  fx <- make_fixture(n = 300)
  m  <- lprobust(fx$y, fx$x, eval = c(0, 0.5))

  capture.output(s90 <- summary(m, alpha = 0.10))
  z90 <- qnorm(1 - 0.10 / 2)
  expect_equal(s90$CI_l, s90$Estimate[, "tau.bc"] - z90 * s90$Estimate[, "se.rb"])
  expect_equal(s90$alpha, 0.10)
})

test_that("cov.us / cov.rb are NULL when covgrid = FALSE", {
  fx <- make_fixture(n = 300)
  m  <- lprobust(fx$y, fx$x, eval = c(0, 0.25, 0.5))
  expect_null(m$cov.us)
  expect_null(m$cov.rb)
})

## For deriv=k, lprobust should return k! * beta_k from the local polynomial.

test_that("lprobust derivative estimates match deriv! * local poly coefficient", {
  fx <- make_fixture()
  kernel <- "uni"
  h <- 0.3
  eval <- 0

  for (deriv in 0:2) {
    for (p in (deriv + 1):(deriv + 2)) {
      est <- lprobust(fx$y, fx$x, eval = eval, h = h, p = p, deriv = deriv,
                      kernel = kernel, vce = "hc0")

      sub <- abs((fx$x - eval) / h) <= 1
      xc <- fx$x[sub] - eval
      ys <- fx$y[sub]
      X <- cbind(1, poly(xc, degree = p, raw = TRUE))
      beta <- solve(crossprod(X), crossprod(X, ys))
      tau_ref <- factorial(deriv) * beta[deriv + 1]

      expect_equal(as.numeric(est$Estimate[, "tau.us"]),
                   as.numeric(tau_ref),
                   tolerance = 1e-10,
                   label = sprintf("deriv=%d p=%d", deriv, p))
    }
  }
})

## Point estimates from lprobust must match the intercept of the equivalent
## weighted least squares fit (uniform kernel = OLS on bandwidth subset).

test_that("lprobust point estimates match lm / WLS across kernels, p, eval", {
  fx <- make_fixture()

  for (kernel in c("uni", "tri", "epa")) {
    for (p in 1:3) {
      for (eval in c(-0.3, 0, 0.4)) {
        h <- 0.25
        est <- lprobust(fx$y, fx$x, eval = eval, h = h, p = p,
                        kernel = kernel, vce = "hc0")

        u  <- (fx$x - eval) / h
        w  <- w_fun(u, kernel)
        sub <- w > 0
        xc <- fx$x[sub] - eval
        ys <- fx$y[sub]
        ws <- w[sub]
        X  <- cbind(1, poly(xc, degree = p, raw = TRUE))
        beta <- solve(crossprod(X * sqrt(ws)), crossprod(X * ws, ys))

        expect_equal(as.numeric(est$Estimate[, "tau.us"]),
                     as.numeric(beta[1]),
                     tolerance = 1e-10,
                     label = sprintf("kernel=%s p=%d eval=%+.1f", kernel, p, eval))
      }
    }
  }
})

## kdrobust point estimate must equal the direct kernel density estimator
## mean(K((x-eval)/h))/h for kernels that have closed-form support.

test_that("kdrobust point estimate matches direct kernel density", {
  set.seed(3)
  xx <- rnorm(2000)
  for (kernel in c("epa", "uni")) {
    for (eval_pt in c(-0.5, 0, 0.7)) {
      h <- 0.35
      kd <- kdrobust(xx, eval = eval_pt, h = h, kernel = kernel)
      tau_ref <- mean(w_fun((xx - eval_pt) / h, kernel)) / h
      expect_equal(as.numeric(kd$Estimate[, "tau.us"]),
                   as.numeric(tau_ref),
                   tolerance = 1e-12,
                   label = sprintf("kernel=%s eval=%+.1f", kernel, eval_pt))
    }
  }
})

test_that("kdrobust rejects unsupported kernels with a clear error", {
  set.seed(3)
  xx <- rnorm(200)
  expect_error(kdrobust(xx, kernel = "tri"), "kernel incorrectly specified")
  expect_error(kdrobust(xx, kernel = "gau"), "kernel incorrectly specified")
})

## broom-style tidy / glance outputs have the expected columns, types,
## and CI interpretation (tau.bc +/- z * se.rb).

skip_if_no_broom <- function() {
  testthat::skip_if_not_installed("broom")
}

test_that("tidy.lprobust returns expected columns", {
  skip_if_no_broom()
  fx <- make_fixture(n = 300)
  m <- lprobust(fx$y, fx$x, eval = c(0, 0.5))
  td <- broom::tidy(m)
  expect_s3_class(td, "data.frame")
  expect_named(td, c("eval","estimate","std.error","tau.bc","se.rb",
                     "h","b","n.eff","conf.low","conf.high"))
  z <- qnorm(0.975)
  expect_equal(td$conf.low,  td$tau.bc - z * td$se.rb, tolerance = 1e-12)
  expect_equal(td$conf.high, td$tau.bc + z * td$se.rb, tolerance = 1e-12)
})

test_that("glance.lprobust returns a one-row summary", {
  skip_if_no_broom()
  fx <- make_fixture(n = 300)
  m <- lprobust(fx$y, fx$x, eval = 0.5)
  gl <- broom::glance(m)
  expect_equal(nrow(gl), 1L)
  expect_true(all(c("n","neval","p","deriv","kernel","bwselect") %in% names(gl)))
})

test_that("tidy/glance on kdrobust and lpbwselect work", {
  skip_if_no_broom()
  fx <- make_fixture(n = 300)
  kd <- kdrobust(fx$x, eval = 0)
  expect_s3_class(broom::tidy(kd), "data.frame")
  expect_equal(nrow(broom::glance(kd)), 1L)

  bw <- lpbwselect(fx$y, fx$x, eval = 0)
  expect_s3_class(broom::tidy(bw), "data.frame")
  expect_equal(nrow(broom::glance(bw)), 1L)
})

## masspoints="check" warns when too few unique x values fall inside the
## bandwidth window. masspoints="off" silences the warning.

test_that("masspoints='check' warns on discrete running variable", {
  set.seed(5)
  # discrete grid with few unique values
  x <- sample(c(-1, -0.5, 0, 0.5, 1), 300, replace = TRUE)
  y <- x + rnorm(300, sd = 0.1)
  expect_warning(
    lprobust(y, x, eval = 0, h = 0.75, p = 1, kernel = "uni",
             vce = "hc0", masspoints = "check"),
    "unique x values within bandwidth"
  )
})

test_that("masspoints='off' suppresses the warning", {
  set.seed(5)
  x <- sample(c(-1, -0.5, 0, 0.5, 1), 300, replace = TRUE)
  y <- x + rnorm(300, sd = 0.1)
  expect_silent(
    lprobust(y, x, eval = 0, h = 0.75, p = 1, kernel = "uni",
             vce = "hc0", masspoints = "off")
  )
})

test_that("masspoints rejects invalid values", {
  fx <- make_fixture(n = 100)
  expect_error(
    lprobust(fx$y, fx$x, eval = 0, h = 0.3, masspoints = "bogus"),
    "masspoints must be one of"
  )
})

test_that("masspoints='check' stays silent with abundant unique x", {
  fx <- make_fixture(n = 800)
  expect_silent(
    lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "uni", vce = "hc0")
  )
})

## All bandwidth selectors must produce finite, strictly positive bandwidths.

test_that("all lpbwselect methods produce finite positive bandwidths", {
  fx <- make_fixture()
  for (bw in c("mse-dpi", "mse-rot", "imse-dpi", "imse-rot", "ce-dpi", "ce-rot")) {
    b <- lpbwselect(fx$y, fx$x, eval = 0, bwselect = bw)
    hh <- b$bws[, "h"]; bb <- b$bws[, "b"]
    expect_true(all(is.finite(hh) & hh > 0), label = paste("h for", bw))
    expect_true(all(is.finite(bb) & bb > 0), label = paste("b for", bw))
  }
})

test_that("interior=TRUE and interior=FALSE both succeed", {
  fx <- make_fixture()
  b1 <- lpbwselect(fx$y, fx$x, eval = 0, bwselect = "mse-dpi", interior = TRUE)
  b2 <- lpbwselect(fx$y, fx$x, eval = 0, bwselect = "mse-dpi", interior = FALSE)
  expect_true(is.finite(b1$bws[, "h"]))
  expect_true(is.finite(b2$bws[, "h"]))
})

test_that("bwcheck > N triggers a warning (bug #8 regression)", {
  fx <- make_fixture(n = 10)
  expect_warning(
    lprobust(fx$y, fx$x, eval = 0.5, bwcheck = 50),
    "bwcheck .* is larger than the sample size"
  )
})

test_that("rho=0 with manual h and no b is an error (bug #5 regression)", {
  fx <- make_fixture()
  expect_error(
    lprobust(fx$y, fx$x, eval = 0.5, h = 0.1, rho = 0),
    "b must also be provided"
  )
})

## User weights should multiply kernel weights throughout: point estimates
## must match a weighted least squares fit with w = K(u) * user_weights.

test_that("weights arg matches WLS with kernel * user weights", {
  fx <- make_fixture(n = 600)
  set.seed(3)
  w <- runif(fx$n, 0.2, 2)

  h <- 0.3; eval <- 0; p <- 1
  est <- lprobust(fx$y, fx$x, eval = eval, h = h, p = p,
                  kernel = "uni", vce = "hc0", weights = w)

  u <- (fx$x - eval) / h
  kw <- w_fun(u, "uni")
  comb <- kw * w
  sub  <- comb > 0
  xc <- fx$x[sub] - eval
  ys <- fx$y[sub]
  ws <- comb[sub]
  X <- cbind(1, xc)
  beta <- solve(crossprod(X * sqrt(ws)), crossprod(X * ws, ys))

  expect_equal(as.numeric(est$Estimate[, "tau.us"]),
               as.numeric(beta[1]),
               tolerance = 1e-10)
})

test_that("weights=1 matches no weights", {
  fx <- make_fixture(n = 400)
  est0 <- lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "epa", vce = "hc0")
  est1 <- lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "epa", vce = "hc0",
                   weights = rep(1, fx$n))
  expect_equal(as.numeric(est0$Estimate), as.numeric(est1$Estimate),
               tolerance = 1e-12)
})

test_that("zero weights exclude observations", {
  fx <- make_fixture(n = 400)
  w <- rep(1, fx$n); w[fx$x > 0.5] <- 0
  est_w <- lprobust(fx$y, fx$x, eval = 0.3, h = 0.3, p = 1, kernel = "uni",
                    vce = "hc0", weights = w)
  est_s <- lprobust(fx$y, fx$x, eval = 0.3, h = 0.3, p = 1, kernel = "uni",
                    vce = "hc0", subset = fx$x <= 0.5)
  expect_equal(as.numeric(est_w$Estimate[, "tau.us"]),
               as.numeric(est_s$Estimate[, "tau.us"]),
               tolerance = 1e-12)
})

## covgrid=TRUE should return symmetric covariance matrices whose diagonal
## equals se.us^2 / se.rb^2 at each eval point.

test_that("covgrid diagonal matches squared SEs", {
  fx <- make_fixture(n = 400)
  est <- lprobust(fx$y, fx$x, eval = seq(0.2, 0.8, length.out = 5),
                  covgrid = TRUE)

  expect_equal(dim(est$cov.us), c(5, 5))
  expect_equal(dim(est$cov.rb), c(5, 5))
  expect_true(isSymmetric(est$cov.us))
  expect_true(isSymmetric(est$cov.rb))

  expect_equal(diag(est$cov.us), as.numeric(est$Estimate[, "se.us"]^2),
               tolerance = 1e-10)
  expect_equal(diag(est$cov.rb), as.numeric(est$Estimate[, "se.rb"]^2),
               tolerance = 1e-10)
})

test_that("covgrid scaling is correct for deriv >= 2", {
  ## Previously V.us.i had factorial(deriv)^2 baked in, so
  ## (V.us.i %*% t(V.us.j))[d+1,d+1] picked up factorial(deriv)^4 instead
  ## of factorial(deriv)^2 -- inflating cov entries by factorial(deriv)^2
  ## for deriv >= 2 (factor 4 at deriv=2; 36 at deriv=3).
  fx <- make_fixture(n = 600)
  for (d in c(0, 1, 2, 3)) {
    est <- lprobust(fx$y, fx$x, p = d + 1, deriv = d,
                    eval = c(0.2, 0.5, 0.7), covgrid = TRUE)
    expect_equal(diag(est$cov.us), as.numeric(est$Estimate[, "se.us"]^2),
                 tolerance = 1e-10,
                 label = sprintf("deriv=%d cov.us diag", d))
    expect_equal(diag(est$cov.rb), as.numeric(est$Estimate[, "se.rb"]^2),
                 tolerance = 1e-10,
                 label = sprintf("deriv=%d cov.rb diag", d))
  }
})

## HC0/HC1/HC2/HC3 standard errors must match the standard sandwich formulas
## evaluated at the corresponding WLS fit.

sandwich_se <- function(y, x, eval, h, vce, p = 1) {
  u <- (x - eval) / h
  sub <- abs(u) <= 1
  xc <- x[sub] - eval
  ys <- y[sub]
  W <- rep(0.5 / h, sum(sub))
  X <- cbind(1, xc)
  XWX_inv <- solve(crossprod(X * sqrt(W)))
  beta <- XWX_inv %*% crossprod(X * W, ys)
  resid <- ys - X %*% beta

  H_mat <- X %*% XWX_inv %*% t(X * W)
  hii <- diag(H_mat)

  nk <- nrow(X); kk <- ncol(X)
  adj <- switch(vce,
                "hc0" = rep(1, nk),
                "hc1" = rep(sqrt(nk / (nk - (p + 1))), nk),
                "hc2" = sqrt(1 / (1 - hii)),
                "hc3" = 1 / (1 - hii))
  e <- as.numeric(adj * resid)
  B_mat <- crossprod(X * W * e, X * W * e)
  vc <- XWX_inv %*% B_mat %*% XWX_inv
  sqrt(vc[1, 1])
}

test_that("HC0/HC1/HC2/HC3 SE match the manual sandwich", {
  fx <- make_fixture()
  for (vce in c("hc0", "hc1", "hc2", "hc3")) {
    est <- lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "uni", vce = vce)
    expect_equal(as.numeric(est$Estimate[, "se.us"]),
                 sandwich_se(fx$y, fx$x, eval = 0, h = 0.3, vce = vce),
                 tolerance = 1e-12,
                 label = paste("vce=", vce))
  }
})

## Cluster-robust variance tests. vce + cluster combinations map to:
##   vce="hc0" + cluster  -> CR0  (raw residuals, no DOF multiplier)
##   vce="hc1" + cluster  -> CR1  (raw residuals, ((n-1)/(n-k))*G/(G-1))
##   vce="hc2" + cluster  -> CR2  (Bell-McCaffrey block-adjusted residuals)
##   vce="hc3" + cluster  -> CR3  (block jackknife (I-H_gg)^{-1}, (G-1)/G multiplier)

cluster_se_manual <- function(y, x, eval, h, cluster, cr_type, p = 1) {
  u   <- (x - eval) / h
  sub <- abs(u) <= 1
  xc  <- x[sub]; ys <- y[sub]; cs <- cluster[sub]
  W   <- rep(0.5 / h, sum(sub))
  sqrtW <- sqrt(W)
  X   <- cbind(1, xc)
  X_std <- X * sqrtW

  XWX_inv <- solve(crossprod(X_std))
  beta <- XWX_inv %*% crossprod(X * W, ys)
  r_raw <- as.numeric(ys - X %*% beta)
  r_std <- r_raw * sqrtW

  clusters <- unique(cs); G <- length(clusters)
  nk <- nrow(X); kk <- ncol(X)

  M <- matrix(0, kk, kk)
  for (g in clusters) {
    ind <- cs == g
    Xg <- X_std[ind, , drop = FALSE]
    rg <- r_std[ind]
    if (cr_type %in% c("CR0", "CR1")) {
      r_adj <- rg
    } else {
      H_gg <- Xg %*% XWX_inv %*% t(Xg)
      H_gg <- (H_gg + t(H_gg)) / 2
      I_H <- diag(nrow(Xg)) - H_gg
      if (cr_type == "CR2") {
        ev <- eigen(I_H, symmetric = TRUE)
        vals <- pmax(ev$values, 1e-12)
        r_adj <- as.numeric(ev$vectors %*% (t(ev$vectors) %*% rg / sqrt(vals)))
      } else {
        r_adj <- drop(solve(I_H, rg))
      }
    }
    score <- crossprod(Xg, r_adj)
    M <- M + tcrossprod(score)
  }
  mult <- switch(cr_type,
                 "CR0" = 1,
                 "CR1" = ((nk - 1) / (nk - kk)) * (G / (G - 1)),
                 "CR2" = 1,
                 "CR3" = (G - 1) / G)
  vc <- XWX_inv %*% (mult * M) %*% XWX_inv
  sqrt(vc[1, 1])
}

test_that("cluster variance: CR1, CR2, CR3 match manual", {
  fx <- make_fixture()
  set.seed(11)
  cl <- sample(1:15, fx$n, replace = TRUE)

  spec <- list(
    list(vce = "cr1", cr = "CR1"),
    list(vce = "cr2", cr = "CR2"),
    list(vce = "cr3", cr = "CR3")
  )
  for (s in spec) {
    est <- lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "uni",
                    vce = s$vce, cluster = cl)
    se_ref <- cluster_se_manual(fx$y, fx$x, eval = 0, h = 0.3, cluster = cl,
                                cr_type = s$cr)
    expect_equal(as.numeric(est$Estimate[, "se.us"]),
                 se_ref, tolerance = 1e-10,
                 label = paste("vce=", s$vce, "->", s$cr))
  }
})

test_that("cluster=1:n with cr1 reduces to HC1-style SE (singleton-cluster sanity)", {
  fx <- make_fixture(n = 200)
  cl <- seq_len(fx$n)   # every obs in its own cluster
  est_cr <- lprobust(fx$y, fx$x, eval = 0, h = 0.3, p = 1, kernel = "uni",
                     vce = "cr1", cluster = cl)
  ## With G == n singleton clusters and CR1's ((n-1)/(n-k)) * (G/(G-1)) =
  ## ((n-1)/(n-k)) * (n/(n-1)) = n/(n-k), the cluster sandwich collapses to
  ## the HC1-style SE on the un-degenerate cluster expansion. Compare to
  ## the manual CR1 reference at G=n.
  se_ref <- cluster_se_manual(fx$y, fx$x, eval = 0, h = 0.3, cluster = cl,
                              cr_type = "CR1")
  expect_equal(as.numeric(est_cr$Estimate[, "se.us"]),
               se_ref, tolerance = 1e-10)
})

lpbwselect = function(y, x, eval=NULL, neval=NULL, p=NULL, deriv=NULL, kernel="epa",
                      bwselect="mse-dpi", bwcheck=21, bwregul=1, imsegrid=30, vce="nn", cluster = NULL,
                      nnmatch=3, interior=FALSE, subset=NULL,
                      weights = NULL, masspoints = "check", data = NULL){

  if (!is.null(data)) {
    mc <- match.call()
    caller_env <- parent.frame()
    .lookup <- function(arg) {
      expr <- mc[[arg]]
      if (is.null(expr)) return(NULL)
      eval(expr, envir = data, enclos = caller_env)
    }
    y <- .lookup("y")
    x <- .lookup("x")
    if ("cluster" %in% names(mc)) cluster <- .lookup("cluster")
    if ("weights" %in% names(mc)) weights <- .lookup("weights")
    if ("subset"  %in% names(mc)) subset  <- .lookup("subset")
  }

  if (!is.null(subset)) {
    x <- x[subset]
    y <- y[subset]
    if (!is.null(cluster)) cluster <- cluster[subset]
    if (!is.null(weights)) weights <- weights[subset]
  }

  ## UX prechecks: catch invalid bwcheck/imsegrid/weights with clear messages
  .nperrs <- character()
  if (!is.null(bwcheck) && (!is.numeric(bwcheck) || length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck <= 0))
    .nperrs <- c(.nperrs, "bwcheck must be a single positive finite integer.")
  if (!is.numeric(imsegrid) || length(imsegrid) != 1 || !is.finite(imsegrid) || imsegrid <= 0)
    .nperrs <- c(.nperrs, "imsegrid must be a single positive finite integer.")
  if (!is.null(weights)) {
    if (!is.numeric(weights))               .nperrs <- c(.nperrs, "weights must be numeric.")
    if (length(weights) != length(x))       .nperrs <- c(.nperrs, "weights length must equal length(x) (after subset).")
    if (sum(weights, na.rm = TRUE) <= 0)    .nperrs <- c(.nperrs, "weights must have a strictly positive sum.")
  }
  if (length(.nperrs) > 0) {
    for (.m in .nperrs) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  na.ok <- complete.cases(x) & complete.cases(y)
  if (!is.null(cluster)) na.ok <- na.ok & complete.cases(cluster)
  if (!is.null(weights)) na.ok <- na.ok & complete.cases(weights) & weights >= 0

  x <- x[na.ok]
  y <- y[na.ok]
  if (!is.null(cluster)) cluster <- cluster[na.ok]
  if (!is.null(weights)) weights <- weights[na.ok]
  if (is.null(weights))  weights <- rep(1, length(x))

  if (!(masspoints %in% c("check", "off"))) {
    stop("masspoints must be one of \"check\" or \"off\".")
  }


  x.max <- max(x); x.min <- min(x)
  N <- length(x)

  if (!is.null(bwcheck)) {
    if (bwcheck > N) {
      warning("bwcheck (", bwcheck, ") is larger than the sample size (", N,
              "); reducing bwcheck to N.")
      bwcheck <- N
    }
  }

  if (!is.null(deriv) & is.null(p)) p <- deriv+1
  if (is.null(p))         p <- 1
  if (is.null(deriv)) deriv <- 0
  q <- p+1

  if (is.null(eval)) {
    if (is.null(neval)) {
      #eval <- unique(x)
      #qseq <- seq(0,1,1/(20+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      eval <- seq(x.min, x.max, length.out=30)
    }
    else {
      #eval <- seq(x.min,x.max,length.out=neval)
      #qseq <- seq(0,1,1/(neval+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      eval <- seq(x.min, x.max, length.out=neval)
    }
  }
  neval <- length(eval)

  if  (bwselect=="imse-dpi" | bwselect=="imse-rot") neval =  eval = 1

  even <- (p-deriv)%%2==0

  kernel   <- tolower(kernel)
  bwselect <- tolower(bwselect)
  vce      <- tolower(vce)

  # Validate vce against the allowed set; mirror lprobust() so that direct
  # calls to lpbwselect() get the same UX as the bw selector invoked from
  # inside lprobust().
  valid_vce <- c("nn", "hc0", "hc1", "hc2", "hc3", "cr1", "cr2", "cr3", "")
  if (!(vce %in% valid_vce))
    stop(sprintf("vce incorrectly specified (received '%s'); allowed: nn, hc0, hc1, hc2, hc3, cr1, cr2, cr3.", vce), call. = FALSE)

  # Cluster vce normalization: with cluster, only cr1/cr2/cr3 are valid;
  # without cluster, cr1/cr2/cr3 fall back to hc1/hc2/hc3. Internal helpers
  # only know about nn/hc0/hc1/hc2/hc3, so cr* gets remapped to hc*.
  if (!is.null(cluster)) {
    if (vce %in% c("nn", "")) {
      vce <- "cr1"  # silent default
    } else if (vce %in% c("hc0", "hc1")) {
      warning(paste0("vce='", vce, "' is not a cluster option. Switching to vce='cr1'."), call. = FALSE)
      vce <- "cr1"
    } else if (vce == "hc2") {
      warning("vce='hc2' is not a cluster option. Switching to vce='cr2'.", call. = FALSE)
      vce <- "cr2"
    } else if (vce == "hc3") {
      warning("vce='hc3' is not a cluster option. Switching to vce='cr3'.", call. = FALSE)
      vce <- "cr3"
    }
  } else {
    if (vce == "cr1") {
      warning("vce='cr1' requires a cluster variable. Falling back to vce='hc1'.", call. = FALSE)
      vce <- "hc1"
    } else if (vce == "cr2") {
      warning("vce='cr2' requires a cluster variable. Falling back to vce='hc2'.", call. = FALSE)
      vce <- "hc2"
    } else if (vce == "cr3") {
      warning("vce='cr3' requires a cluster variable. Falling back to vce='hc3'.", call. = FALSE)
      vce <- "hc3"
    }
  }
  if (vce == "cr1") vce <- "hc1"
  if (vce == "cr2") vce <- "hc2"
  if (vce == "cr3") vce <- "hc3"

  kernel.type <- "Epanechnikov"
  if (kernel=="triangular"   | kernel=="tri") kernel.type <- "Triangular"
  if (kernel=="uniform"      | kernel=="uni") kernel.type <- "Uniform"
  if (kernel=="gaussian"     | kernel=="gau") kernel.type <- "Gaussian"

  bws           <- matrix(NA,neval,2)
  colnames(bws) <- c("h","b")
  bws.imse      <- NULL

  if (bwselect=="all") {
    bws           <- matrix(NA,neval,8)
    colnames(bws) <- c("h.mse.dpi","b.mse.dpi", "h.mse.rot","b.mse.rot", "h.ce.dpi","b.ce.dpi", "h.ce.rot","b.ce.rot")
    bws.imse      <- matrix(NA,2,2)
  }

  if  (bwselect=="imse-dpi" | bwselect=="all") {
      est <- lpbwselect.imse.dpi(y=y, x=x, cluster=cluster, weights=weights,
                                 p=p, q=q, deriv=deriv, kernel=kernel,
                                 bwcheck=bwcheck, bwregul=bwregul,
                                 imsegrid=imsegrid, vce=vce,
                                 nnmatch=nnmatch, interior=interior)
      h.imse.dpi   <- est$h
      b.imse.dpi   <- est$b
      bws[1,1:2]   <- c(h.imse.dpi,  b.imse.dpi)
  }

  if  (bwselect=="imse-rot" | bwselect=="all") {
      est <- lpbwselect.imse.rot(y=y, x=x, p=p, deriv=deriv, kernel=kernel, imsegrid=imsegrid)
      h.imse.rot   <- est$h
      est <- lpbwselect.imse.rot(y=y, x=x, p=q, deriv=p+1,   kernel=kernel, imsegrid=imsegrid)
      b.imse.rot   <- est$h
      bws[1,1:2]   <- c(h.imse.rot,  b.imse.rot)
  }

  if  (bwselect=="all") {
    bws.imse[,1] <- c(h.imse.dpi,  b.imse.dpi)
    bws.imse[,2] <- c(h.imse.rot,  b.imse.rot)
  }

  if  (bwselect=="all"  | bwselect=="mse-dpi" | bwselect=="mse-rot" | bwselect=="ce-dpi" | bwselect=="ce-rot") {

     for (i in 1:neval) {

      if  (bwselect=="mse-dpi" | bwselect=="ce-dpi" | bwselect=="ce-rot" | bwselect=="all") {
        est <- lpbwselect.mse.dpi(y=y, x=x, cluster=cluster, weights=weights,
                                  eval=eval[i], p=p, q=q, deriv=deriv,
                                  kernel=kernel, bwcheck=bwcheck,
                                  bwregul=bwregul, vce=vce, nnmatch=nnmatch,
                                  interior=interior)
        h.mse.dpi  <- est$h
        b.mse.dpi  <- est$b
        bws[i,1:2] <- c(h.mse.dpi,  b.mse.dpi)
      }

      if  (bwselect=="mse-rot" | bwselect=="all") {
        est <- lpbwselect.mse.rot(y=y, x=x, eval=eval[i], p=p, deriv=deriv, kernel=kernel)
        h.mse.rot  <- est$h
        est <- lpbwselect.mse.rot(y=y, x=x, eval=eval[i], p=q, deriv=p+1,   kernel=kernel)
        b.mse.rot  <- est$h
        bws[i,1:2] <- c(h.mse.rot,  b.mse.rot)
      }

      h.ce.dpi=b.ce.dpi=h.ce.rot=b.ce.rot=0

      if  (bwselect=="ce-dpi" | bwselect=="all") {
        h.ce.dpi=b.ce.dpi=0
        #if (deriv==0) {
          if  (isTRUE(even) ) {
            h.ce.dpi <- h.mse.dpi*N^(-((p+2)/((2*p+5)*(p+3))))
            b.ce.dpi <- b.mse.dpi*N^(-((q)/((2*q+3)*(q+3))))
          } else{
            est <- lpbwselect.ce.dpi(y=y, x=x, h=h.mse.dpi, b=b.mse.dpi,
                                     eval=eval[i], p=p, q=q, deriv=deriv,
                                     rho=1, kernel=kernel, vce=vce,
                                     nnmatch=nnmatch, interior=interior,
                                     bwregul=bwregul, weights=weights)
            h.ce.dpi <- est$h
            b.ce.dpi <- b.mse.dpi*N^(-((q+2)/((2*q+5)*(q+3))))
          }
            bws[i,1:2] <- c(h.ce.dpi,  b.ce.dpi)
        #}
      }

      if  (bwselect=="ce-rot" | bwselect=="all") {
        h.ce.rot=b.ce.rot=0
        #if (deriv==0) {
          if  (isTRUE(even) ) {
            h.ce.rot <- h.mse.dpi*N^(-((p+2)/((2*p+5)*(p+3))))
            b.ce.rot <- b.mse.dpi*N^(-((q)/((2*q+3)*(q+3))))
          } else{
            h.ce.rot <- h.mse.dpi*N^(-((p)/((2*p+3)*(p+3))))
            b.ce.rot <- b.mse.dpi*N^(-((q+2)/((2*q+5)*(q+3))))
          }
          bws[i,1:2]  <- c(h.ce.rot,  b.ce.rot)
      #}
      }

      if (bwselect=="all") bws[i,] <- c(h.mse.dpi,b.mse.dpi, h.mse.rot,b.mse.rot,  h.ce.dpi,b.ce.dpi, h.ce.rot,b.ce.rot)
    }
  }

  bws <- cbind(eval, bws)
  out        <- list(bws = bws, bws.imse = bws.imse,
                     opt = list(n=N, neval=neval, p=p, q=q, deriv=deriv, kernel=kernel.type, bwselect=bwselect))
  out$call   <- match.call()
  class(out) <- "lpbwselect"
  return(out)
}

print.lpbwselect <- function(x,...){
  cat("Call: lpbwselect\n\n")

  cat(paste("Sample size (n)                              =    ", x$opt$n,        "\n", sep=""))
  cat(paste("Polynomial order for point estimation (p)    =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Order of derivative estimated (deriv)        =    ", x$opt$deriv,    "\n", sep=""))
  cat(paste("Polynomial order for confidence interval (q) =    ", x$opt$q,        "\n", sep=""))
  cat(paste("Kernel function                              =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                             =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  invisible(x)
}

summary.lpbwselect <- function(object, sep = 5, ...) {
  out <- list(opt      = object$opt,
              bws      = object$bws,
              bws.imse = object$bws.imse,
              sep      = sep)
  class(out) <- "summary.lpbwselect"
  out
}

print.summary.lpbwselect <- function(x, ...) {
  cat("Call: lpbwselect\n\n")

  cat(paste("Sample size (n)                              =    ", x$opt$n,        "\n", sep=""))
  cat(paste("Polynomial order for point estimation (p)    =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Order of derivative estimated (deriv)        =    ", x$opt$deriv,    "\n", sep=""))
  cat(paste("Polynomial order for confidence interval (q) =    ", x$opt$q,        "\n", sep=""))
  cat(paste("Kernel function                              =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                             =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  sep <- x$sep

  if (x$opt$bwselect=="all") {
    col1.names = c("","", "MSE-DPI","", "MSE-ROT", "","CE-DPI", "","CE-ROT")
    col2.names = rep(c("h", "b"),4)
  } else {
    col1.names = c("")
    col2.names = c("h", "b")
  }

  ### print output
  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
  }
  if (x$opt$bwselect=="all") {
    cat(format(col1.names  , width=8, justify="right"))
    cat("\n")
  }
  if (x$opt$bwselect!="imse-dpi" & x$opt$bwselect!="imse-rot") cat(format("eval", width=10, justify="right"))
  cat(format(col2.names            , width=8, justify="right"))
  cat("\n")

  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
  }

  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(format(sprintf("%3.3f", x$bws[2:3])  , width=9, justify="right"))
    cat("\n")
  } else {
    for (j in 1:nrow(x$bws)) {
      cat(format(toString(j), width=4))
      cat(format(sprintf("%3.3f", x$bws[j, "eval"]), width=8, justify="right"))
      cat(format(sprintf("%3.3f", x$bws[j, 2:ncol(x$bws)])  , width=8, justify="right"))
      cat("\n")
      if (is.numeric(sep)) if (sep > 0) if (j %% sep == 0) {
        cat(paste(rep("-", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
      }
    }
  }

  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
  }

  if (x$opt$bwselect=="all") {
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat(format(c("","IMSE-DPI","", "IMSE-ROT") , width=8, justify="right"))
    cat("\n")
    cat(format(c("h", "b", "h", "b")            , width=8, justify="right"))
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat(format(sprintf("%3.3f", x$bws.imse)  , width=8, justify="right"))
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat("\n")
  }

  invisible(x)
}

## Tests for cluster-vce surface: cr1/cr2/cr3 acceptance, nn+cluster
## redirects to cr1, hc-cluster aliases match cr* equivalents, and cr*
## without cluster falls back to hc* with a warning.

make_cluster_fixture <- function(n = 800, n_clusters = 30, seed = 11) {
  set.seed(seed)
  x  <- sort(runif(n, -1, 1))
  y  <- 1 + 2 * x + 3 * x^2 + rnorm(n, sd = 0.4)
  cl <- ceiling(n_clusters * runif(n))
  list(x = x, y = y, cl = cl)
}

test_that("cr1/cr2/cr3 require cluster (without cluster they warn + fall back)", {
  d <- make_cluster_fixture()
  expect_warning(
    m1 <- lprobust(d$y, d$x, eval = 0, vce = "cr1"),
    "requires a cluster variable"
  )
  expect_warning(
    m2 <- lprobust(d$y, d$x, eval = 0, vce = "cr2"),
    "requires a cluster variable"
  )
  expect_warning(
    m3 <- lprobust(d$y, d$x, eval = 0, vce = "cr3"),
    "requires a cluster variable"
  )
  ## After fallback: cr1->hc1, cr2->hc2, cr3->hc3 produces the same
  ## numbers as if the user had passed hc1/hc2/hc3 directly.
  m_hc1 <- lprobust(d$y, d$x, eval = 0, vce = "hc1")
  m_hc2 <- lprobust(d$y, d$x, eval = 0, vce = "hc2")
  m_hc3 <- lprobust(d$y, d$x, eval = 0, vce = "hc3")
  expect_equal(m1$Estimate[, "se.us"], m_hc1$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m2$Estimate[, "se.us"], m_hc2$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m3$Estimate[, "se.us"], m_hc3$Estimate[, "se.us"], tolerance = 1e-12)
})

test_that("vce='nn' + cluster silently defaults to cr1", {
  d <- make_cluster_fixture()
  ## No warning expected for the silent default.
  expect_silent(
    m_default <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "nn")
  )
  ## Same as explicit cr1.
  m_cr1 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr1")
  expect_equal(m_default$Estimate[, "se.us"], m_cr1$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m_default$Estimate[, "se.rb"], m_cr1$Estimate[, "se.rb"], tolerance = 1e-12)
})

test_that("hc{0,1,2,3} + cluster warns and remaps to cr1/cr1/cr2/cr3", {
  d <- make_cluster_fixture()
  expect_warning(
    m_hc0 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "hc0"),
    "Switching to vce='cr1'"
  )
  expect_warning(
    m_hc1 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "hc1"),
    "Switching to vce='cr1'"
  )
  expect_warning(
    m_hc2 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "hc2"),
    "Switching to vce='cr2'"
  )
  expect_warning(
    m_hc3 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "hc3"),
    "Switching to vce='cr3'"
  )
  m_cr1 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr1")
  m_cr2 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr2")
  m_cr3 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr3")
  expect_equal(m_hc0$Estimate[, "se.us"], m_cr1$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m_hc1$Estimate[, "se.us"], m_cr1$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m_hc2$Estimate[, "se.us"], m_cr2$Estimate[, "se.us"], tolerance = 1e-12)
  expect_equal(m_hc3$Estimate[, "se.us"], m_cr3$Estimate[, "se.us"], tolerance = 1e-12)
})

test_that("CR1, CR2, CR3 produce different SEs (sanity)", {
  d <- make_cluster_fixture()
  m1 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr1")
  m2 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr2")
  m3 <- lprobust(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr3")
  expect_false(isTRUE(all.equal(m1$Estimate[, "se.us"], m2$Estimate[, "se.us"])))
  expect_false(isTRUE(all.equal(m1$Estimate[, "se.us"], m3$Estimate[, "se.us"])))
  expect_false(isTRUE(all.equal(m2$Estimate[, "se.us"], m3$Estimate[, "se.us"])))
  ## CR3 >= CR1 elementwise (block jackknife is conservative)
  expect_true(m3$Estimate[, "se.us"] >= m1$Estimate[, "se.us"])
})

test_that("invalid vce names error", {
  d <- make_cluster_fixture()
  expect_error(
    suppressWarnings(lprobust(d$y, d$x, eval = 0, vce = "garbage")),
    "vce incorrectly specified|invalid input"
  )
})

test_that("lpbwselect mirrors lprobust vce surface", {
  d <- make_cluster_fixture()
  ## lpbwselect with cr1 (cluster supplied) returns same h as hc1 (which
  ## also gets remapped to cr1 with a warning).
  expect_silent(
    bw_cr1 <- lpbwselect(d$y, d$x, eval = 0, cluster = d$cl, vce = "cr1")
  )
  expect_warning(
    bw_hc1 <- lpbwselect(d$y, d$x, eval = 0, cluster = d$cl, vce = "hc1"),
    "Switching to vce='cr1'"
  )
  expect_equal(bw_cr1$bws[, "h"], bw_hc1$bws[, "h"], tolerance = 1e-12)
  expect_equal(bw_cr1$bws[, "b"], bw_hc1$bws[, "b"], tolerance = 1e-12)
})

lprobust = function(y, x, eval=NULL, neval=NULL, p=NULL, deriv=NULL, h=NULL, b=NULL, rho=1,
                    kernel="epa", bwselect=NULL, bwcheck=21, bwregul=1, imsegrid=30, vce="nn", covgrid = FALSE,
                    cluster=NULL, nnmatch=3, level=95, interior = FALSE, subset = NULL,
                    weights = NULL, masspoints = "check", data = NULL) {

  if (!is.null(data)) {
    mc <- match.call()
    caller_env <- parent.frame()
    .lookup <- function(arg) {
      expr <- mc[[arg]]
      if (is.null(expr)) return(NULL)
      eval(expr, envir = data, enclos = caller_env)
    }
    y <- .lookup("y")
    x <- .lookup("x")
    if ("cluster" %in% names(mc)) cluster <- .lookup("cluster")
    if ("weights" %in% names(mc)) weights <- .lookup("weights")
    if ("subset"  %in% names(mc)) subset  <- .lookup("subset")
  }

  ## Validate aux-vector lengths against length(x) BEFORE subset filtering,
  ## so wrong-length inputs error explicitly instead of silently recycling.
  .n_orig <- length(x)
  if (length(y) != .n_orig)
    stop(sprintf("'y' and 'x' must have equal length (got y=%d, x=%d).", length(y), .n_orig), call. = FALSE)
  if (!is.null(cluster) && length(cluster) != .n_orig)
    stop(sprintf("'cluster' must have length equal to length(x) (got %d, expected %d).", length(cluster), .n_orig), call. = FALSE)
  if (!is.null(weights) && length(weights) != .n_orig)
    stop(sprintf("'weights' must have length equal to length(x) (got %d, expected %d).", length(weights), .n_orig), call. = FALSE)
  if (!is.null(subset)) {
    if (is.logical(subset)) {
      if (length(subset) != .n_orig)
        stop(sprintf("Logical 'subset' must have length equal to length(x) (got %d, expected %d).", length(subset), .n_orig), call. = FALSE)
    } else if (is.numeric(subset)) {
      if (any(!is.finite(subset)) || any(subset < 1) || any(subset > .n_orig) || any(subset != round(subset)))
        stop(sprintf("Numeric 'subset' must contain integer indices in 1..%d.", .n_orig), call. = FALSE)
    } else {
      stop("'subset' must be logical or integer.", call. = FALSE)
    }
  }

  if (!is.null(subset)) {
    x <- x[subset]
    y <- y[subset]
    if (!is.null(cluster)) cluster <- cluster[subset]
    if (!is.null(weights)) weights <- weights[subset]
    if (length(x) == 0L) stop("'subset' removed all observations.", call. = FALSE)
  }

  ## UX prechecks: catch invalid h/b/bwcheck/imsegrid/weights/cluster/eval/level/rho/etc.
  .nperrs <- character()
  .bwbad <- function(v, name) {
    if (is.null(v))         return(NULL)
    if (!is.numeric(v))     return(paste0(name, " must be numeric."))
    if (any(!is.finite(v))) return(paste0(name, " contains non-finite values (NA/Inf)."))
    if (any(v <= 0))        return(paste0(name, " must be strictly positive."))
    NULL
  }
  .nperrs <- c(.nperrs, .bwbad(h, "h"), .bwbad(b, "b"))
  if (!is.null(bwcheck) && (!is.numeric(bwcheck) || length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck <= 0 || bwcheck != round(bwcheck)))
    .nperrs <- c(.nperrs, "bwcheck must be a single positive integer.")
  if (!is.numeric(imsegrid) || length(imsegrid) != 1 || !is.finite(imsegrid) || imsegrid <= 0 || imsegrid != round(imsegrid))
    .nperrs <- c(.nperrs, "imsegrid must be a single positive integer.")
  if (!is.numeric(level) || length(level) != 1 || !is.finite(level) || level <= 0 || level >= 100)
    .nperrs <- c(.nperrs, "level must be a single number in (0, 100).")
  if (!is.numeric(rho) || length(rho) != 1 || !is.finite(rho) || rho < 0)
    .nperrs <- c(.nperrs, "rho must be a single non-negative number.")
  if (!is.null(eval)) {
    if (!is.numeric(eval) || any(!is.finite(eval)))
      .nperrs <- c(.nperrs, "eval must be numeric and finite.")
    if (length(eval) == 0L)
      .nperrs <- c(.nperrs, "eval must have at least one element.")
  }
  if (!is.null(weights)) {
    if (!is.numeric(weights))               .nperrs <- c(.nperrs, "weights must be numeric.")
    if (length(weights) != length(x))       .nperrs <- c(.nperrs, "weights length must equal length(x) (after subset).")
    if (sum(weights, na.rm = TRUE) <= 0)    .nperrs <- c(.nperrs, "weights must have a strictly positive sum.")
  }
  if (length(.nperrs) > 0) {
    for (.m in .nperrs) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  na.ok <- complete.cases(x) & complete.cases(y)
  if (!is.null(cluster)) na.ok <- na.ok & complete.cases(cluster)
  if (!is.null(weights)) na.ok <- na.ok & complete.cases(weights) & weights >= 0

  x <- x[na.ok]
  y <- y[na.ok]
  if (!is.null(cluster)) cluster <- cluster[na.ok]
  if (!is.null(weights)) weights <- weights[na.ok]
  if (is.null(weights))  weights <- rep(1, length(x))
  
  if (!is.null(deriv) & is.null(p)) p <- deriv+1
  if (is.null(p))         p <- 1
  if (is.null(deriv)) deriv <- 0
  q <- p+1
  
  x.max <- max(x); x.min <- min(x)
  N <- length(x)

  if (!is.null(bwcheck)) {
    if (bwcheck > N) {
      warning("bwcheck (", bwcheck, ") is larger than the sample size (", N,
              "); reducing bwcheck to N.")
      bwcheck <- N
    }
  }

  if (is.null(eval)) {
    if (is.null(neval)) {
      #eval <- unique(x)
      #qseq <- seq(0,1,1/(20+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      eval <- seq(x.min, x.max, length.out=30)
    }
    else {
      #eval <- seq(x.min,x.max,length.out=neval)
      #qseq <- seq(0,1,1/(neval+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      eval <- seq(x.min, x.max, length.out=neval)
    }
  }
  neval <- length(eval)

  ## Precheck (continued): h/b length must be 1 or neval, otherwise indexing later silently recycles
  .nperrs2 <- character()
  if (!is.null(h) && length(h) != 1L && length(h) != neval)
    .nperrs2 <- c(.nperrs2, paste0("h must have length 1 or neval (=", neval, ")."))
  if (!is.null(b) && length(b) != 1L && length(b) != neval)
    .nperrs2 <- c(.nperrs2, paste0("b must have length 1 or neval (=", neval, ")."))
  if (length(.nperrs2) > 0) {
    for (.m in .nperrs2) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  if (is.null(h) & is.null(bwselect) & neval==1) bwselect="mse-dpi"
  if (is.null(h) & is.null(bwselect) & neval>1)  bwselect="imse-dpi"
  
  if (vce=="nn") {
    order.x <- order(x)
    x <- x[order.x]
    y <- y[order.x]
    if (!is.null(cluster)) cluster <- cluster[order.x]
    weights <- weights[order.x]
  }

  if (!(masspoints %in% c("check", "off"))) {
    stop("masspoints must be one of \"check\" or \"off\".")
  }

  kernel   <- tolower(kernel)
  bwselect <- tolower(bwselect)
  vce      <- tolower(vce)

  #####################################################   CHECK ERRORS
  exit <- 0
  .err <- function(msg) { warning(msg, call. = FALSE); exit <<- 1 }

  valid_kernels  <- c("gau","gaussian","uni","uniform","tri","triangular","epa","epanechnikov","")
  valid_vce      <- c("nn","hc0","hc1","hc2","hc3","cr1","cr2","cr3","")
  valid_bwselect <- c("mse-dpi","mse-rot","imse-dpi","imse-rot","ce-dpi","ce-rot","manual","")

  if (!(kernel %in% valid_kernels)) .err("kernel incorrectly specified")
  if (!(vce    %in% valid_vce))     .err("vce incorrectly specified")
  ## bwselect may be character(0) (after tolower(NULL)) when h is supplied
  ## manually; in that case it gets set to "Manual" later. Skip check then.
  if (!is.null(bwselect) && length(bwselect) > 0L && nzchar(bwselect) &&
      !(bwselect %in% valid_bwselect)) {
    if (identical(bwselect, "all")) {
      .err("bwselect=\"all\" is only supported by lpbwselect; lprobust requires a single method (e.g. \"mse-dpi\", \"imse-dpi\", \"ce-dpi\").")
    } else {
      .err(paste0("bwselect incorrectly specified (received '", bwselect, "')"))
    }
  }

  if (p < 0 || deriv < 0 || nnmatch <= 0)
    .err("p, deriv must be >=0 and nnmatch must be >0")
  if (deriv > p)
    .err("deriv must be <= p")
  ## integer roundness check
  if (p > 0 && round(p) != p)             .err("p must be an integer")
  if (deriv >= 0 && round(deriv) != deriv) .err("deriv must be an integer")
  if (nnmatch > 0 && round(nnmatch) != nnmatch) .err("nnmatch must be an integer")

  if (exit > 0) stop("nprobust: invalid input (see warnings above).", call. = FALSE)

  # Cluster vce validation: with cluster, only cr1/cr2/cr3 are valid.
  # Without cluster, cr1/cr2/cr3 fall back to hc1/hc2/hc3.
  if (!is.null(cluster)) {
    if (vce %in% c("nn", "")) {
      vce <- "cr1"  # silent default
    } else if (vce %in% c("hc0", "hc1")) {
      warning(paste0("vce='", vce, "' is not a cluster option. Switching to vce='cr1'."), call. = FALSE)
      vce <- "cr1"
    } else if (vce == "hc2") {
      warning("vce='hc2' is not a cluster option. Switching to vce='cr2'.", call. = FALSE)
      vce <- "cr2"
    } else if (vce == "hc3") {
      warning("vce='hc3' is not a cluster option. Switching to vce='cr3'.", call. = FALSE)
      vce <- "cr3"
    }
  } else {
    if (vce == "cr1") {
      warning("vce='cr1' requires a cluster variable. Falling back to vce='hc1'.", call. = FALSE)
      vce <- "hc1"
    } else if (vce == "cr2") {
      warning("vce='cr2' requires a cluster variable. Falling back to vce='hc2'.", call. = FALSE)
      vce <- "hc2"
    } else if (vce == "cr3") {
      warning("vce='cr3' requires a cluster variable. Falling back to vce='hc3'.", call. = FALSE)
      vce <- "hc3"
    }
  }

  # User-facing display label.
  vce_type <- "NN"
  if (vce == "hc0") vce_type <- "HC0"
  if (vce == "hc1") vce_type <- "HC1"
  if (vce == "hc2") vce_type <- "HC2"
  if (vce == "hc3") vce_type <- "HC3"
  if (vce == "cr1") vce_type <- "CR1"
  if (vce == "cr2") vce_type <- "CR2"
  if (vce == "cr3") vce_type <- "CR3"

    if (!is.null(h)) bwselect <- "Manual"
    if (!is.null(h) & rho>0 & is.null(b)) {
        b <- h/rho
    }
    if (!is.null(h) & rho==0 & is.null(b)) {
        stop("When h is provided and rho=0, b must also be provided (b cannot be computed as h/rho).")
    }
    
  kernel.type <- "Epanechnikov"
  if (kernel=="triangular"   | kernel=="tri") kernel.type <- "Triangular"
  if (kernel=="uniform"      | kernel=="uni") kernel.type <- "Uniform"
  if (kernel=="gaussian"     | kernel=="gau") kernel.type <- "Gaussian"

    
  ############################################################################################
    #print("Preparing data.")
    if (is.null(h)) {
        lpbws <- lpbwselect(y=y, x=x,  eval=eval, deriv=deriv, p=p, vce=vce,
                            cluster=cluster, bwselect=bwselect, interior=interior,
                            kernel=kernel, bwcheck=bwcheck, bwregul=bwregul,
                            imsegrid=imsegrid, subset=NULL,
                            weights=weights, masspoints="off")
        h     <- lpbws$bws[,2]
        b     <- lpbws$bws[,3]
        if (rho>0) b <- h/rho
        rho   <- h/b
    }

  # Internal mapping: cr1/cr2/cr3 -> hc1/hc2/hc3. Done AFTER the lpbwselect
  # call so that lpbwselect receives the user-facing cr* label and can do
  # its own (warning-free) normalization without double-firing the
  # cluster-vce warning the user has already seen.
  if (vce == "cr1") vce <- "hc1"
  if (vce == "cr2") vce <- "hc2"
  if (vce == "cr3") vce <- "hc3"
  
    if (length(h)==1 & neval>1) {
      h <- rep(h,neval)
      b <- rep(b,neval)
      rho   <- h/b
    }
  
  dups <- dupsid <- 0
  if (vce=="nn") {
    # x is sorted at this point; rle() gives run-length encoding of duplicates.
    runs <- rle(x)
    dups   <- rep.int(runs$lengths, runs$lengths)
    dupsid <- sequence(runs$lengths)
  }
  
  cov.p = NULL
  ####################################################
  Estimate=matrix(NA,neval,8)
  colnames(Estimate)=c("eval","h","b","N","tau.us","tau.bc","se.us","se.rb")
  
  
  for (i in 1:neval) {
    
    if (!is.null(bwcheck)) {
      bw.min   <- sort(abs(x-eval[i]))[bwcheck]
      #nh       <- sum(abs(x-eval[i]) <= h)
      #nb       <- sum(abs(x-eval[i]) <= b)
      h[i]     <- max(h[i], bw.min)
      b[i]     <- max(b[i], bw.min)
    }
    
    w.h   <- W.fun((x-eval[i])/h[i], kernel)/h[i] * weights
    w.b   <- W.fun((x-eval[i])/b[i], kernel)/b[i] * weights
    ind.h <- w.h>0;  ind.b <- w.b>0
    N.h   <- sum(ind.h);  N.b <- sum(ind.b)
    ind   <- ind.b
    if (h[i]>b[i]) ind <- ind.h

    if (masspoints == "check") {
      n_unique <- length(unique(x[ind.h]))
      if (n_unique < (p + 5)) {
        warning(sprintf(
          "Only %d unique x values within bandwidth at eval=%.4f (p+5=%d); local polynomial may be unreliable. Set masspoints=\"off\" to silence.",
          n_unique, eval[i], p + 5))
      }
    }
    
    #if (N.h.l<5 | N.h.r<5 | N.b.l<5 | N.b.r<5){
    #  stop("Not enough observations to perform calculations")
    #  exit(1)
    #}
    
    eN  <- sum(ind)
    eY  <- y[ind]
    eX  <- x[ind]
    W.h <- w.h[ind]
    W.b <- w.b[ind]
    
    eC = NULL
    if (!is.null(cluster)) eC = cluster[ind]
    
    edups <- edupsid <- 0	
    if (vce=="nn") {
      edups   <- dups[ind]
      edupsid <- dupsid[ind]
    #  for (j in 1:eN) {
    #    edups[j]=sum(eX==eX[j])
    #  }
    #  j=1
    #  while (j<=eN) {
    #    edupsid[j:(j+edups[j]-1)] <- 1:edups[j]
    #    j <- j+edups[j]
    #  }
    }
    
  u   <- (eX-eval[i])/h[i]
  R.q <- matrix(NA,eN,(q+1))
  for (j in 1:(q+1))  R.q[,j] <- (eX-eval[i])^(j-1)
  R.p <- R.q[,1:(p+1)]

  #display("Computing RD estimates.")
  L <- crossprod(R.p*W.h,u^(p+1)) 
  invG.q  <- qrXXinv((sqrt(W.b)*R.q))
  invG.p  <- qrXXinv((sqrt(W.h)*R.p))
  e.p1    <- matrix(0,(q+1),1); e.p1[p+2]=1
  e.v     <- matrix(0,(p+1),1); e.v[deriv+1]=1
  Q.q     <- t(t(R.p*W.h) - h[i]^(p+1)*(L%*%t(e.p1))%*%t(t(invG.q%*%t(R.q))*W.b))
  beta.p  <- invG.p%*%crossprod(R.p*W.h,eY); beta.q <- invG.q%*%crossprod(R.q*W.b,eY); beta.bc <- invG.p%*%crossprod(Q.q,eY) 

  tau.cl <- factorial(deriv)*beta.p[(deriv+1),1]
  tau.bc <- factorial(deriv)*beta.bc[(deriv+1),1]

  hii <- predicts.p <- predicts.q <- 0
  if (vce %in% c("hc0","hc1","hc2","hc3")) {
    predicts.p <- R.p %*% beta.p
    predicts.q <- R.q %*% beta.q
    if (vce %in% c("hc2","hc3") && is.null(eC)) {
      # vectorized diag(R %*% invG %*% t(R*W)) -- row-wise dot product
      hii <- matrix(rowSums((R.p %*% invG.p) * (R.p * W.h)), eN, 1)
    }
  }

  if (is.null(eC)) {
    res.h <- lprobust.res(eX, eY, predicts.p, hii, vce, nnmatch, edups, edupsid, p+1)
    res.b <- if (vce == "nn") res.h else lprobust.res(eX, eY, predicts.q, hii, vce, nnmatch, edups, edupsid, q+1)
    V.Y.cl <- invG.p %*% lprobust.vce(as.matrix(R.p * W.h), res.h, NULL) %*% invG.p
    V.Y.bc <- invG.p %*% lprobust.vce(Q.q, res.b, NULL) %*% invG.p
  } else {
    cr_type <- switch(vce,
                      "hc0" = "CR0",
                      "hc1" = "CR1",
                      "hc2" = "CR2",
                      "hc3" = "CR3",
                      "nn"  = "CR1",
                      "CR1")
    if (vce == "nn") {
      res.h.raw <- lprobust.res(eX, eY, predicts.p, hii, "nn", nnmatch, edups, edupsid, p+1)
      res.b.raw <- res.h.raw
    } else {
      res.h.raw <- matrix(as.numeric(eY) - as.numeric(predicts.p), ncol = 1)
      res.b.raw <- matrix(as.numeric(eY) - as.numeric(predicts.q), ncol = 1)
    }
    sqrtW.h  <- sqrt(W.h)
    X.std.h  <- R.p * sqrtW.h
    r.std.h  <- as.numeric(res.h.raw) * sqrtW.h
    # k_override = q+1 aligns the CR1 df correction with the q-regression
    # that produced res.b.raw. Without it, k=ncol(Q.q)=p+1 was being used.
    meat.cl  <- lprobust.cluster.meat(X.std.h, r.std.h, eC, invG.p, cr_type)
    meat.bc  <- lprobust.cluster.meat(Q.q, as.numeric(res.b.raw), eC, invG.p, cr_type, k_override = q + 1)
    V.Y.cl   <- invG.p %*% meat.cl %*% invG.p
    V.Y.bc   <- invG.p %*% meat.bc %*% invG.p
  }
	se.cl  <- sqrt(factorial(deriv)^2*V.Y.cl[deriv+1,deriv+1])
	se.rb  <- sqrt(factorial(deriv)^2*V.Y.bc[deriv+1,deriv+1])
	
	Estimate[i,] <- c(eval[i], h[i], b[i], eN, tau.cl, tau.bc, se.cl, se.rb) 
  }
  
  cov.us <- NULL
  cov.rb <- NULL

  if (isTRUE(covgrid)) {

    cov.us <- matrix(NA, neval, neval)
    cov.rb <- matrix(NA, neval, neval)

    ## Loop-invariants hoisted out of the (i, j) inner loop.
    e.p1     <- matrix(0, q + 1, 1); e.p1[p + 2]  <- 1
    e.v      <- matrix(0, p + 1, 1); e.v[deriv+1] <- 1
    fact_d   <- factorial(deriv)
    is_hc    <- vce %in% c("hc0", "hc1", "hc2", "hc3")
    is_h2or3 <- vce %in% c("hc2", "hc3")

    ## Diagonal: cov.us[i,i] = se.us[i]^2 (and similarly for cov.rb)
    ## by construction. Avoids neval redundant inner-loop iterations.
    diag(cov.us) <- as.numeric(Estimate[, "se.us"])^2
    diag(cov.rb) <- as.numeric(Estimate[, "se.rb"])^2

    if (neval > 1) for (i in 1:(neval - 1)) {
      for (j in (i + 1):neval) {

        w.h.i   <- W.fun((x-eval[i])/h[i], kernel)/h[i]
        w.b.i   <- W.fun((x-eval[i])/b[i], kernel)/b[i]
        ind.h.i <- w.h.i>0;  ind.b.i <- w.b.i>0
        ind.i   <- ind.h.i | ind.b.i

        w.h.j   <- W.fun((x-eval[j])/h[j], kernel)/h[j]
        w.b.j   <- W.fun((x-eval[j])/b[j], kernel)/b[j]
        ind.h.j <- w.h.j>0;  ind.b.j <- w.b.j>0
        ind.j   <- ind.h.j | ind.b.j

        ind <- ind.i | ind.j

        eN  <- sum(ind)
        eY  <- y[ind]
        eX  <- x[ind]

        W.h.i <- w.h.i[ind]
        W.b.i <- w.b.i[ind]
        W.h.j <- w.h.j[ind]
        W.b.j <- w.b.j[ind]

        eC <- if (!is.null(cluster)) cluster[ind] else NULL

        edups <- edupsid <- 0
        if (vce == "nn") {
          edups   <- dups[ind]
          edupsid <- dupsid[ind]
        }

        u.i   <- (eX-eval[i])/h[i]
        R.q.i <- outer(eX - eval[i], 0:q, `^`)
        R.p.i <- R.q.i[, 1:(p+1), drop = FALSE]

        u.j   <- (eX-eval[j])/h[j]
        R.q.j <- outer(eX - eval[j], 0:q, `^`)
        R.p.j <- R.q.j[, 1:(p+1), drop = FALSE]

        L.i       <- crossprod(R.p.i*W.h.i, u.i^(p+1))
        invG.q.i  <- qrXXinv((sqrt(W.b.i)*R.q.i))
        invG.p.i  <- qrXXinv((sqrt(W.h.i)*R.p.i))
        Q.q.i     <- t(t(R.p.i*W.h.i) - h[i]^(p+1)*(L.i%*%t(e.p1))%*%t(t(invG.q.i%*%t(R.q.i))*W.b.i))
        beta.p.i  <- invG.p.i%*%crossprod(R.p.i*W.h.i, eY)
        beta.q.i  <- invG.q.i%*%crossprod(R.q.i*W.b.i, eY)

        L.j       <- crossprod(R.p.j*W.h.j, u.j^(p+1))
        invG.q.j  <- qrXXinv((sqrt(W.b.j)*R.q.j))
        invG.p.j  <- qrXXinv((sqrt(W.h.j)*R.p.j))
        Q.q.j     <- t(t(R.p.j*W.h.j) - h[j]^(p+1)*(L.j%*%t(e.p1))%*%t(t(invG.q.j%*%t(R.q.j))*W.b.j))
        beta.p.j  <- invG.p.j%*%crossprod(R.p.j*W.h.j, eY)
        beta.q.j  <- invG.q.j%*%crossprod(R.q.j*W.b.j, eY)

        hii.i <- predicts.p.i <- predicts.q.i <- 0
        hii.j <- predicts.p.j <- predicts.q.j <- 0
        if (is_hc) {
          predicts.p.i <- R.p.i %*% beta.p.i
          predicts.q.i <- R.q.i %*% beta.q.i
          predicts.p.j <- R.p.j %*% beta.p.j
          predicts.q.j <- R.q.j %*% beta.q.j
          if (is_h2or3) {
            hii.i <- matrix(rowSums((R.p.i %*% invG.p.i) * (R.p.i * W.h.i)), eN, 1)
            hii.j <- matrix(rowSums((R.p.j %*% invG.p.j) * (R.p.j * W.h.j)), eN, 1)
          }
        }

        res.h.i <- lprobust.res(eX, eY, predicts.p.i, hii.i, vce, nnmatch, edups, edupsid, p+1)
        res.b.i <- if (vce == "nn") res.h.i
                   else lprobust.res(eX, eY, predicts.q.i, hii.i, vce, nnmatch, edups, edupsid, q+1)

        res.h.j <- lprobust.res(eX, eY, predicts.p.j, hii.j, vce, nnmatch, edups, edupsid, p+1)
        res.b.j <- if (vce == "nn") res.h.j
                   else lprobust.res(eX, eY, predicts.q.j, hii.j, vce, nnmatch, edups, edupsid, q+1)

        ## NOTE: V.us.i carries factorial(deriv) (not factorial(deriv)^2) so
        ## that cov(tau_i, tau_j) = (V.us.i %*% t(V.us.j))[d+1,d+1] picks up
        ## factorial(deriv)^2 from the cross-product -- matching the scaling
        ## of se.us = sqrt(factorial(deriv)^2 * V.Y.cl[d+1,d+1]) in the main
        ## loop. The previous implementation had factorial(deriv)^2 here,
        ## which gave factorial(deriv)^4 after the cross-product (a factor
        ## factorial(deriv)^2 too large for deriv >= 2).
        V.us.i <- fact_d * invG.p.i %*% t(c(res.h.i) * R.p.i * W.h.i)
        V.us.j <- fact_d * invG.p.j %*% t(c(res.h.j) * R.p.j * W.h.j)

        V.rb.i <- fact_d * invG.p.i %*% t(c(res.b.i) * Q.q.i)
        V.rb.j <- fact_d * invG.p.j %*% t(c(res.b.j) * Q.q.j)

        cov.us[i, j] <- (V.us.i %*% t(V.us.j))[deriv+1, deriv+1]
        cov.rb[i, j] <- (V.rb.i %*% t(V.rb.j))[deriv+1, deriv+1]

        cov.us[j, i] <- cov.us[i, j]
        cov.rb[j, i] <- cov.rb[i, j]
      }
    }

  }
    
    
  out        <-list(Estimate=Estimate, opt=list(p=p, q=q, deriv=deriv, kernel=kernel.type, n=N, neval=neval, bwselect=bwselect), cov.us=cov.us, cov.rb=cov.rb)
  out$call   <- match.call()
  class(out) <- "lprobust"
  return(out)
}

print.lprobust <- function(x,...){
  cat("Call: lprobust\n\n")

  cat(paste("Sample size (n)                              =    ", x$opt$n,        "\n", sep=""))
  cat(paste("Polynomial order for point estimation (p)    =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Order of derivative estimated (deriv)        =    ", x$opt$deriv,    "\n", sep=""))
  cat(paste("Polynomial order for confidence interval (q) =    ", x$opt$q,        "\n", sep=""))
  cat(paste("Kernel function                              =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                             =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  invisible(x)
}

summary.lprobust <- function(object, alpha = 0.05, sep = 5, ...) {
  z    <- qnorm(1 - alpha / 2)
  CI_l <- object$Estimate[, "tau.bc"] - object$Estimate[, "se.rb"] * z
  CI_r <- object$Estimate[, "tau.bc"] + object$Estimate[, "se.rb"] * z

  out <- list(opt      = object$opt,
              Estimate = object$Estimate,
              alpha    = alpha,
              sep      = sep,
              CI_l     = CI_l,
              CI_r     = CI_r)
  class(out) <- "summary.lprobust"
  out
}

print.summary.lprobust <- function(x, ...) {
  cat("Call: lprobust\n\n")

  cat(paste("Sample size (n)                              =    ", x$opt$n,        "\n", sep=""))
  cat(paste("Polynomial order for point estimation (p)    =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Order of derivative estimated (deriv)        =    ", x$opt$deriv,    "\n", sep=""))
  cat(paste("Polynomial order for confidence interval (q) =    ", x$opt$q,        "\n", sep=""))
  cat(paste("Kernel function                              =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                             =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  alpha <- x$alpha
  sep   <- x$sep

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  cat(format(" ", width= 14 ))
  cat(format(" ", width= 10 ))
  cat(format(" ", width= 8  ))
  cat(format("Point", width= 10, justify="right"))
  cat(format("Std." , width= 10, justify="right"))
  cat(format("Robust B.C.", width=25, justify="centre"))
  cat("\n")

  cat(format("eval"            , width=14, justify="right"))
  cat(format("h"               , width=10, justify="right"))
  cat(format("Eff.n"           , width=8 , justify="right"))
  cat(format("Est."            , width=10, justify="right"))
  cat(format("Error"           , width=10, justify="right"))
  cat(format(paste("[ ", floor((1-alpha)*100), "%", " C.I. ]", sep=""), width=25, justify="centre"))
  cat("\n")

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  for (j in 1:nrow(x$Estimate)) {
    cat(format(toString(j), width=4))
    cat(format(sprintf("%3.3f", x$Estimate[j, "eval"]), width=10, justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[j, "h"])  , width=10, justify="right"))
    cat(format(sprintf("%3.0f", x$Estimate[j, "N"])  , width=8 , justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[j, "tau.us"]) , width=10, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$Estimate[j, "se.us"]), sep=""), width=10, justify="right"))
    cat(format(paste("[", sprintf("%3.3f", x$CI_l[j]), " , ", sep="")  , width=14, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$CI_r[j]), "]", sep=""), width=11, justify="left"))
    cat("\n")
    if (is.numeric(sep)) if (sep > 0) if (j %% sep == 0) {
      cat(paste(rep("-", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")
    }
  }

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  invisible(x)
}

## Regression tests for input-validation hardening (2026-05-08).
library(testthat)
library(nprobust)

set.seed(20260508)
n  <- 400
x  <- runif(n, -1, 1)
y  <- 0.3 * x + rnorm(n)
cl <- ceiling(15 * runif(n))

# Helper that captures warnings + errors so the warning-then-stop pattern
# (warning() in .err(); then stop("nprobust: invalid input...")) is caught.
expect_input_error <- function(expr, pattern) {
  warns <- character(0)
  out <- tryCatch(
    withCallingHandlers(
      expr,
      warning = function(w) { warns <<- c(warns, conditionMessage(w)); invokeRestart("muffleWarning") }
    ),
    error = function(e) e$message
  )
  all_text <- c(if (is.character(out)) out else character(0), warns)
  expect_true(any(grepl(pattern, all_text, fixed = TRUE)),
              info = paste0("expected match='", pattern, "'; got: ",
                            paste(all_text, collapse = " | ")))
}

test_that("level must be in (0, 100)", {
  expect_input_error(lprobust(y, x, level = 100),
                     "level must be a single number")
  expect_input_error(lprobust(y, x, level = 0),
                     "level must be a single number")
  expect_input_error(kdrobust(x, level = 100),
                     "level must be a single number")
  expect_input_error(kdrobust(x, level = c(95, 90)),
                     "level must be a single number")
})

test_that("rho must be a finite non-negative scalar (kdrobust)", {
  expect_input_error(kdrobust(x, rho = -1),
                     "rho must be a single non-negative number")
  expect_input_error(kdrobust(x, rho = c(1, 2)),
                     "rho must be a single non-negative number")
})

test_that("eval finiteness validated in kdrobust / kdbwselect", {
  expect_input_error(kdrobust(x, eval = c(0, NA)),
                     "eval must be numeric and finite")
  expect_input_error(kdbwselect(x, eval = c(0, Inf)),
                     "eval must be numeric and finite")
})

test_that("bwselect=\"all\" rejected by lprobust / kdrobust with clear message", {
  expect_input_error(lprobust(y, x, bwselect = "all"),
                     "only supported by lpbwselect")
  expect_input_error(kdrobust(x, bwselect = "all"),
                     "only supported by kdbwselect")
})

test_that("bwcheck must be a positive integer", {
  expect_input_error(lprobust(y, x, bwcheck = -5),
                     "bwcheck must be a single positive integer")
  expect_input_error(lprobust(y, x, bwcheck = 2.5),
                     "bwcheck must be a single positive integer")
})

test_that("vector lengths validated against length(x)", {
  expect_input_error(lprobust(y, x, weights = rep(1, 10)),
                     "must have length equal to length(x)")
  expect_input_error(lprobust(y, x, cluster = cl[1:10]),
                     "must have length equal to length(x)")
  expect_input_error(lprobust(y, x, subset = rep(TRUE, 10)),
                     "Logical 'subset' must have length")
})

test_that("empty subset errors", {
  expect_input_error(lprobust(y, x, subset = rep(FALSE, n)),
                     "removed all observations")
})

test_that("eval must be finite", {
  expect_input_error(lprobust(y, x, eval = c(0.0, NA, 0.5)),
                     "eval must be numeric and finite")
})

test_that("bwselect typo rejected", {
  expect_input_error(lprobust(y, x, bwselect = "no-such-method"),
                     "bwselect incorrectly specified")
})

test_that("manual h with rho=0 and no b works (regression)", {
  # rho=0 + manual h + manual b is OK (b is provided directly)
  expect_silent(suppressWarnings(lprobust(y, x, eval = 0, h = 0.5, b = 0.5, rho = 0)))
})

## Regression test for the CR1 cluster + bias-corrected k_override fix
## (2026-05-08). Pre-fix the cluster meat for V.Y.bc used k=ncol(Q.q)=p+1
## in the (N-1)/(N-k) df correction; the residuals come from the q-regression
## so the correct k is q+1. lm + sandwich::vcovCL with q-regressor design is
## the unambiguous anchor.

library(testthat)
library(nprobust)

test_that("CR1 cluster bias-corrected SE matches lm(p=q=2)+vcovCL anchor at h=b", {
  skip_if_not_installed("sandwich")
  set.seed(20260508)
  n  <- 1500
  x  <- runif(n, -1, 1)
  y  <- 0.5 * x + 0.3 * x^2 + rnorm(n)
  cl <- ceiling(20 * runif(n))

  eval_pt <- 0
  h       <- 0.5

  fit <- lprobust(y, x, eval = eval_pt, h = h, b = h, p = 1, deriv = 0,
                  kernel = "tri", vce = "cr1", cluster = cl)
  se_np <- as.numeric(fit$Estimate[, "se.rb"])

  ## Manual lm anchor with triangular kernel weights and q-regression design
  in_bw <- abs(x - eval_pt) <= h
  yi <- y[in_bw]; xi <- x[in_bw]; cli <- cl[in_bw]
  u  <- (xi - eval_pt) / h
  wi <- (1 - abs(u)) * (abs(u) <= 1)
  nz <- wi > 0
  yi <- yi[nz]; xi <- xi[nz]; cli <- cli[nz]; wi <- wi[nz]
  sw <- sqrt(wi)
  xc <- xi - eval_pt
  Xq <- cbind(1, xc, xc^2)
  fit_lm <- lm(I(sw * yi) ~ I(sw * Xq) - 1)
  V_q    <- sandwich::vcovCL(fit_lm, cluster = cli, type = "HC1")
  se_lm  <- sqrt(V_q[1, 1])

  expect_lt(abs(se_np - se_lm), 1e-10)
})

test_that("CR2 / CR3 cluster paths unchanged by k_override fix", {
  set.seed(123)
  n  <- 800
  x  <- runif(n, -1, 1)
  y  <- 0.4 * x + rnorm(n)
  cl <- ceiling(15 * runif(n))

  for (vce in c("cr2", "cr3")) {
    fit <- lprobust(y, x, eval = 0, h = 0.5, b = 0.5, p = 1, vce = vce, cluster = cl)
    expect_true(is.finite(fit$Estimate[, "se.rb"]))
    expect_gt(fit$Estimate[, "se.rb"], 0)
  }
})

## broom-compatible tidy()/glance() methods.
##
## Registered in NAMESPACE via `S3method(broom::tidy, <class>)` so they
## dispatch when broom is loaded, without requiring broom at install time.
##
## Convention:
##   - `estimate`   = tau.us  (classical/conventional point estimate)
##   - `std.error`  = se.us   (its SE)
##   - `conf.low`/`conf.high` = robust bias-corrected CI:
##        tau.bc +/- z * se.rb
##   - additional columns tau.bc, se.rb, h, b, n.eff are kept for
##     downstream convenience.

tidy.lprobust <- function(x, conf.int = TRUE, conf.level = 0.95, ...) {
  est <- as.data.frame(x$Estimate)
  out <- data.frame(
    eval       = est$eval,
    estimate   = est$tau.us,
    std.error  = est$se.us,
    tau.bc     = est$tau.bc,
    se.rb      = est$se.rb,
    h          = est$h,
    b          = est$b,
    n.eff      = est$N,
    row.names  = NULL,
    stringsAsFactors = FALSE
  )
  if (isTRUE(conf.int)) {
    z <- stats::qnorm(1 - (1 - conf.level) / 2)
    out$conf.low  <- est$tau.bc - z * est$se.rb
    out$conf.high <- est$tau.bc + z * est$se.rb
  }
  out
}

glance.lprobust <- function(x, ...) {
  data.frame(
    n        = x$opt$n,
    neval    = x$opt$neval,
    p        = x$opt$p,
    q        = x$opt$q,
    deriv    = x$opt$deriv,
    kernel   = x$opt$kernel,
    bwselect = x$opt$bwselect,
    stringsAsFactors = FALSE
  )
}

tidy.kdrobust <- function(x, conf.int = TRUE, conf.level = 0.95, ...) {
  est <- as.data.frame(x$Estimate)
  out <- data.frame(
    eval       = est$eval,
    estimate   = est$tau.us,
    std.error  = est$se.us,
    tau.bc     = est$tau.bc,
    se.rb      = est$se.rb,
    h          = est$h,
    b          = est$b,
    n.eff      = est$N,
    row.names  = NULL,
    stringsAsFactors = FALSE
  )
  if (isTRUE(conf.int)) {
    z <- stats::qnorm(1 - (1 - conf.level) / 2)
    out$conf.low  <- est$tau.bc - z * est$se.rb
    out$conf.high <- est$tau.bc + z * est$se.rb
  }
  out
}

glance.kdrobust <- function(x, ...) {
  data.frame(
    n        = x$opt$n,
    neval    = x$opt$neval,
    kernel   = x$opt$kernel,
    bwselect = x$opt$bwselect,
    stringsAsFactors = FALSE
  )
}

tidy.lpbwselect <- function(x, ...) {
  bws <- as.data.frame(x$bws)
  rownames(bws) <- NULL
  bws
}

glance.lpbwselect <- function(x, ...) {
  data.frame(
    n        = x$opt$n,
    neval    = x$opt$neval,
    p        = x$opt$p,
    q        = x$opt$q,
    deriv    = x$opt$deriv,
    kernel   = x$opt$kernel,
    bwselect = x$opt$bwselect,
    stringsAsFactors = FALSE
  )
}

tidy.kdbwselect <- function(x, ...) {
  bws <- as.data.frame(x$bws)
  rownames(bws) <- NULL
  bws
}

glance.kdbwselect <- function(x, ...) {
  data.frame(
    n        = x$opt$n,
    neval    = x$opt$neval,
    kernel   = x$opt$kernel,
    bwselect = x$opt$bwselect,
    stringsAsFactors = FALSE
  )
}

kdbwselect <- function(x, eval=NULL, neval=NULL, kernel="epa",
                       bwselect="mse-dpi", bwcheck=21, imsegrid=30, subset = NULL,
                       data = NULL){

  if (!is.null(data)) {
    mc <- match.call()
    caller_env <- parent.frame()
    .lookup <- function(arg) {
      expr <- mc[[arg]]
      if (is.null(expr)) return(NULL)
      eval(expr, envir = data, enclos = caller_env)
    }
    x <- .lookup("x")
    if ("subset" %in% names(mc)) subset <- .lookup("subset")
  }

  p <- 2
  deriv <- 0
  kernel   <- tolower(kernel)
  bwselect <- tolower(bwselect)

  if (!(kernel %in% c("epa","epanechnikov","uni","uniform"))) {
    stop("kernel incorrectly specified. Supported kernels for kdbwselect: epa, uni.")
  }
  if (kernel %in% c("epanechnikov")) kernel <- "epa"
  if (kernel %in% c("uniform"))      kernel <- "uni"

  if (!is.null(subset)) x <- x[subset]

  ## UX prechecks: catch invalid bwcheck/imsegrid with clear messages
  .nperrs <- character()
  if (!is.null(bwcheck) && (!is.numeric(bwcheck) || length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck <= 0))
    .nperrs <- c(.nperrs, "bwcheck must be a single positive finite integer.")
  if (!is.numeric(imsegrid) || length(imsegrid) != 1 || !is.finite(imsegrid) || imsegrid <= 0)
    .nperrs <- c(.nperrs, "imsegrid must be a single positive finite integer.")
  if (!is.null(eval)) {
    if (!is.numeric(eval) || any(!is.finite(eval)))
      .nperrs <- c(.nperrs, "eval must be numeric and finite.")
    if (length(eval) == 0L)
      .nperrs <- c(.nperrs, "eval must have at least one element.")
  }
  if (length(.nperrs) > 0) {
    for (.m in .nperrs) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  na.ok <- complete.cases(x)
  x <- x[na.ok]

  x.min <- min(x);  x.max <- max(x)
  N <- length(x)

  if (!is.null(bwcheck)) {
    if (bwcheck > N) {
      warning("bwcheck (", bwcheck, ") is larger than the sample size (", N,
              "); reducing bwcheck to N.")
      bwcheck <- N
    }
  }

  if (is.null(eval)) {
    if (is.null(neval)) {
      #eval <- unique(x)
      qseq <- seq(0.1,0.9,length.out=30)
      eval <- quantile(x, qseq)
      #eval <- seq(x.min, x.max, length.out=30+2)
      #eval <- eval[2:31]
    }
    else {
      qseq <- seq(0.1,0.9,length.out=neval)
      eval <- quantile(x, qseq)
      #eval <- seq(x.min,x.max,length.out=neval)
      #qseq <- seq(0,1,1/(neval+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      #eval <- seq(x.min, x.max, length.out=(neval+2))
      #eval <- eval[2:(neval+1)]
    }
  }
  neval <- length(eval)


  if (kernel=="epa") {
    kernel.type <- "Epanechnikov"
    C.h <- 2.34
    C.b <- 3.49
  } else {
    kernel.type <- "Uniform"
    C.h <- 1.06
    C.b <- 1
  }
  
  #C.fun <- function(v,r) {
  #  R.v <- kd.K.fun(1, v<-v, r<-r, kernel<-"gau")$R.v
  #  k.v <- kd.K.fun(1, v<-v, r<-0, kernel<-"gau")$k.v
  #  C <- 2*((sqrt(pi)*(1+2*d)*factorial(p+d)*R.v)/(2*p*k.v^2*factorial(2*p+2*d)))^(1/(2*p+2*d+1))
  #  return(C)
  #}

  
  bws           <- matrix(NA,neval,2)
  colnames(bws) <- c("h","b")
  bws.imse <- NULL  
  
  if (bwselect=="all") {
    bws           <- matrix(NA,neval,6)
    colnames(bws) <- c("h.mse.dpi","b.mse.dpi",  "h.ce.dpi","b.ce.dpi", "h.ce.rot","b.ce.rot")
    bws.imse      <- matrix(NA,2,2)
  }

  
  ### IMSE-ROT
  h.imse.rot <- sd(x)*C.h*N^(-1/(1+2*p))
  b.imse.rot <- sd(x)*C.b*N^(-1/(1+2*(p+2)+2*p))
  
  if (bwselect=="imse-rot") {
    bws[,1] <- rep(h.imse.rot, neval)
    bws[,2] <- rep(b.imse.rot, neval)
  }
  
  
  ### IMSE-ROT
  if (bwselect=="imse-dpi" | bwselect=="all") {
  
  B.h=V.h=0  
  #qseq.imse <- seq(0,1,1/(imsegrid+1))
  #eval.imse <- quantile(x, qseq[2:(length(qseq)-1)])
  qseq.imse <- seq(0.1,0.9,length.out=imsegrid)
  eval.imse <- quantile(x, qseq.imse)
  
  #eval.imse <- seq(x.min, x.max, length.out=(imsegrid+2))
  #eval.imse <- eval[2:(imsegrid+1)]
  
  for (i in 1:imsegrid) {
    K.b <- kd.K.fun((x-eval.imse[i])/b.imse.rot, v=p+2, r=p,     kernel=kernel)    
    K.h <- kd.K.fun((x-eval.imse[i])/h.imse.rot, v=p,   r=deriv, kernel=kernel)
    f.b <- mean(K.b$Kx)/b.imse.rot^(1+p)
    f.h.rot <- mean(K.h$Kx)/h.imse.rot    
    B.h[i]  <- f.b*K.h$k.v
    V.h[i]  <- f.h.rot*K.h$R.v
  }
  
  h.imse.dpi <- kd.bw.fun(mean(V.h), mean(B.h), N, v=p, r=deriv)
  
  if (bwselect=="imse-dpi") {
    bws[,1] <- rep(h.imse.dpi, neval)
    bws[,2] <- rep(b.imse.rot, neval)
  }

  }
  
  if  (bwselect=="all") {
    bws.imse[,1] <- c(h.imse.dpi,  b.imse.rot)
    bws.imse[,2] <- c(h.imse.rot,  b.imse.rot)
  }
  
  
  if  (bwselect=="all"  | bwselect=="mse-dpi" | bwselect=="ce-dpi" | bwselect=="ce-rot" ) {
  
    B.h=V.h=0 
    
  for (i in 1:neval) {
  
    #bws[i,1:2] <- c(h.imse.rot, b.imse.rot)
    #q.rot <- sd(x)*C.b*N^(-1/(1+2*(p+4)+2*(p+2)))
    
    #V = (mean((K^2)) - mean(K)^2)/(N*h.rot^2)
    #B = mean(rho^(1+p)*M)/h.rot
    
    #K.q = kd.K.fun((x-eval[i])/q.rot, v=p+4, r=p+2,   kernel=kernel)    
    #f.q.rot = mean(K.q$Kx)/q.rot^(1+2*p)
    #f.q.rot <- kd.K.fun(eval[i], v=2, r=4, kernel="gau")$Kx
    
    #V.b <- f.h.rot*kd.K.fun(1, v=p+2, r=p, kernel=kernel)$R.v
    #B.b <- f.q.rot*kd.K.fun(1, v=p+2, r=p, kernel=kernel)$k.v
    #b.mse.dpi <- bw.fun(V.b, B.b, N, v=p+2, r=p)
    
    if (!is.null(bwcheck)) {
      bw.min     <- sort(abs(x-eval[i]))[bwcheck]
      #h.imse.rot <- max(h.imse.rot, bw.min)
      #b.imse.rot <- max(b.imse.rot, bw.min)
    }
    
    K.b <- kd.K.fun((x-eval[i])/b.imse.rot, v=p+2, r=p,     kernel=kernel)    
    K.h <- kd.K.fun((x-eval[i])/h.imse.rot, v=p,   r=deriv, kernel=kernel)
    f.b <- mean(K.b$Kx)/b.imse.rot^(1+p)
    f.h.rot <- mean(K.h$Kx)/h.imse.rot    
    B.h[i]  <- f.b*K.h$k.v
    V.h[i]  <- f.h.rot*K.h$R.v
    h.mse.dpi <- kd.bw.fun(V.h[i], B.h[i], N, v=p, r=deriv)
    
    #f.h.rot=dnorm(eval[i])
    #f.b.rot=(eval[i]^2-1)*dnorm(eval[i])
    #V.h = (f.h.rot*K.h$v.k)/(N*h.rot^(1+2*deriv))
    #B.h = (f.b.rot*K.h$m.k)*b.rot^p
    #u = (x-eval[i])/h.rot
    #K.d = kd.K.fun(u     , r=p,   d=deriv, kernel=kernel)
    #L.r = kd.K.fun(rho[i]*u , r=p+2, d=p,     kernel=kernel)
    #K = K.d$k.x
    #M = rho[i]^(1+p)*L.r$k.x*L.r$m.k
    #f.hat.us = mean(K)/h.rot
    #bias = mean(M)/h.rot
    #var.us = (mean((K^2)) - mean(K)^2)/(N*h.rot^2)
    #bw.fun(var.us, f.hat.bc, r=p, d=deriv)
    #(var.us/(2*p*N*bias^2))^1/(1+2*p)
    b.mse.dpi = b.imse.rot
    
    if (!is.null(bwcheck)) {
      h.mse.dpi <- max(h.mse.dpi, bw.min)
      b.mse.dpi <- max(b.mse.dpi, bw.min)
    }
    
    bws[i,1:2] <- c(h.mse.dpi,    b.mse.dpi)
    
    if (bwselect=="ce-rot" | bwselect=="all") {
      h.ce.rot <- h.mse.dpi*N^(-(p-2)/((1+2*p)*(1+p+2)))
      b.ce.rot <- b.mse.dpi*N^(-(p-2)/((1+2*p)*(1+p+2)))
      
      if (!is.null(bwcheck)) {
        h.ce.rot <- max(h.ce.rot, bw.min)
        b.ce.rot <- max(b.ce.rot, bw.min)
      }
      
      bws[i,1:2] <- c(h.ce.rot, b.ce.rot)
    }
    
    if (bwselect=="ce-dpi"| bwselect=="all") {
      h.ce.dpi <- kd.cer.fun(x, eval[i], h.mse.dpi, h.mse.dpi, p, kernel)
      b.ce.dpi <- b.mse.dpi
      
      if (!is.null(bwcheck)) {
        h.ce.dpi <- max(h.ce.dpi, bw.min)
        b.ce.dpi <- max(b.ce.dpi, bw.min)
      }
      
      bws[i,1:2] <- c(h.ce.dpi,  b.ce.dpi)
    }
    
    if(bwselect=="all") bws[i,1:6] <- c(h.mse.dpi,b.mse.dpi, h.ce.dpi,b.ce.dpi, h.ce.rot,b.ce.rot)
  }
  }
  
  
  
  bws <- cbind(eval,bws)
  out <- list(bws=bws, bws.imse = bws.imse, opt=list(p=p, n=N, neval=neval, kernel=kernel.type, bwselect=bwselect))
  out$call   <- match.call()
  class(out) <- "kdbwselect"
  return(out)
}

print.kdbwselect <- function(x,...){
  cat("Call: kdbwselect\n\n")

  cat(paste("Sample size (n)                         =    ", x$opt$n,        "\n", sep=""))
#  cat(paste("Kernel order for point estimation (p)   =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Kernel function                         =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                        =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  invisible(x)
}

summary.kdbwselect <- function(object, sep = 5, ...) {
  out <- list(opt      = object$opt,
              bws      = object$bws,
              bws.imse = object$bws.imse,
              sep      = sep)
  class(out) <- "summary.kdbwselect"
  out
}

print.summary.kdbwselect <- function(x, ...) {
  cat("Call: kdbwselect\n\n")

  cat(paste("Sample size (n)                             =    ", x$opt$n,        "\n", sep=""))
#  cat(paste("Polynomial order for point estimation (p)   =    ", x$opt$p,        "\n", sep=""))
  cat(paste("Kernel function                             =    ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth selection method                  =    ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  sep <- x$sep

  if (x$opt$bwselect=="all") {
    col1.names = c("","", "MSE-DPI", "","CE-DPI","","CE-ROT")
    col2.names = rep(c("h", "b"),3)
  } else {
    col1.names = c("")
    col2.names = c("h", "b")
  }
  
  ### print output
  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
  }
  if (x$opt$bwselect=="all") {
    cat(format(col1.names  , width=8, justify="right"))
    cat("\n")
  }
  if (x$opt$bwselect!="imse-dpi" & x$opt$bwselect!="imse-rot") cat(format("eval", width=10, justify="right"))
  cat(format(col2.names            , width=8, justify="right"))
  cat("\n")
  
  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
  }
  
  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(format(sprintf("%3.3f", x$bws[1,2:3])  , width=9, justify="right"))
    cat("\n")
  } else {
    for (j in 1:nrow(x$bws)) {
      cat(format(toString(j), width=4))
      cat(format(sprintf("%3.3f", x$bws[j, "eval"]), width=8, justify="right"))
      cat(format(sprintf("%3.3f", x$bws[j, 2:ncol(x$bws)])  , width=8, justify="right"))
      cat("\n")
      if (is.numeric(sep)) if (sep > 0) if (j %% sep == 0) {
        cat(paste(rep("-", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")
      }
    }
  }
  
  if (x$opt$bwselect=="imse-dpi" | x$opt$bwselect=="imse-rot") {
    cat(paste(rep("=", 15 + 8), collapse="")); cat("\n")
  } else {
    cat(paste(rep("=", 15 + 8*ncol(x$bws)), collapse="")); cat("\n")   
  }
  
  if (x$opt$bwselect=="all") {
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat(format(c("","IMSE-DPI","", "IMSE-ROT") , width=8, justify="right"))
    cat("\n")
    cat(format(c("h", "b", "h", "b")            , width=8, justify="right"))
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat(format(sprintf("%3.3f", x$bws.imse)  , width=8, justify="right"))
    cat("\n")
    cat(paste(rep("=", 15 + 10 + 10), collapse="")); cat("\n")
    cat("\n")
  }

  invisible(x)
}

lpbwce = function(y, x, K, L, res, c, p, q, h, b, deriv, fact) {
  # Pure-R port of the coverage-error DPI helper (formerly Rcpp/Armadillo).
  # Computes the three leading terms q1rbc, q2rbc, q3rbc of the Edgeworth
  # expansion used by lpbwselect.ce.dpi. Equivalent up to machine precision
  # to the original lpbwce.cpp implementation.

  ## Accept y/x/K/L/res as vector or 1-column matrix
  y   <- as.numeric(y)
  x   <- as.numeric(x)
  K   <- as.numeric(K)
  L   <- as.numeric(L)
  res <- as.numeric(res)

  N   <- length(y)
  rho <- h / b

  Xh <- (x - c) / h
  Xb <- (x - c) / b

  Rp <- outer(Xh, 0:p, `^`)
  Rq <- outer(Xb, 0:q, `^`)
  Wp <- K / h
  Wq <- L / b

  Lp1 <- crossprod(Rp, Wp * Xh^(p + 1)) / N
  Gp  <- crossprod(Rp * Wp, Rp) / N
  Gq  <- crossprod(Rq * Wq, Rq) / N
  iGp <- solve(Gp)
  iGq <- solve(Gq)

  ep1 <- numeric(q + 1); ep1[p + 2]    <- 1
  e0  <- numeric(p + 1); e0[deriv + 1] <- fact

  lus0     <- as.numeric((t(e0) %*% iGp) %*% t(Rp * K))
  scale_bc <- as.numeric((t(e0) %*% iGp) %*% Lp1)
  tail_bc  <- as.numeric((t(ep1) %*% iGq) %*% t(Rq * L))
  lbc0 <- lus0 - rho^(p + 1) * scale_bc * tail_bc

  vx <- res^2
  s2 <- sum(lbc0^2 * vx) / (N * h)

  Krrp   <- crossprod(Rp * K, Rp)
  Lrrq   <- crossprod(Rq * L, Rq)
  Krxip  <- crossprod(Rp, K * Xh^(p + 1))
  sumKRp <- crossprod(Rp, K)
  Sxp1   <- sum(Xh^(p + 1))
  Krxp   <- as.numeric(sumKRp) * Sxp1 - as.numeric(Krxip)

  EKrrp      <- Krrp / N
  EKrxip_vec <- as.numeric(Krxip) / N
  EKrxp_vec  <- Krxp / (N * (N - 1))
  ELrrq      <- Lrrq / N

  a_row <- matrix(fact * iGp[deriv + 1, ], 1, p + 1)
  u_a   <- drop(a_row)
  v_a   <- drop(a_row %*% EKrrp %*% iGp)

  RpiGp  <- Rp %*% iGp
  RqiGq  <- Rq %*% iGq
  quadRp <- rowSums(RpiGp * Rp)
  quadRq <- rowSums(RqiGq * Rq)

  Rp_va <- as.numeric(Rp %*% v_a)
  Rp_ua <- as.numeric(Rp %*% u_a)

  ## Single-index sums (vectorized)
  q1  <- sum(lbc0^3 * res^3)
  q3  <- sum(lbc0^4 * (res^4 - vx^2))
  q3a <- q1
  q8  <- sum((lbc0 * res)^4)
  q9  <- sum((lbc0^2 * vx - h * s2) * (lbc0 * res)^2)
  q12 <- sum((lbc0^2 * vx - h * s2)^2)
  q4  <- sum(lbc0^2 * L * quadRq * res^2)

  q5a <- matrix(colSums((lbc0^3 * res^2) * RqiGq), 1)
  q5b <- matrix(colSums((lbc0 * res^2 * L) * Rq), q + 1, 1)
  q7a <- matrix(colSums((lbc0 * res^2 * L) * RqiGq), 1)
  q7b <- crossprod(Rq * lbc0, Rq * lbc0) %*% iGq
  q7c <- matrix(colSums((lbc0 * res^2 * L) * Rq), q + 1, 1)

  ## lbc1 at (i,i) for q2
  lus1_diag <- K * Rp_va - K^2 * Rp_ua * quadRp

  C1_scalar <- drop(a_row %*% EKrrp %*% iGp %*% Lp1)
  v_iGq_ep1 <- drop(iGq %*% ep1)
  C2_vec    <- as.numeric(Rq %*% v_iGq_ep1)
  v_lp      <- drop(iGp %*% Lp1)
  D2_vec    <- as.numeric(Rp %*% v_lp)

  T1_diag <- L * C2_vec * (C1_scalar - K * Rp_ua * D2_vec)

  dot_a_krxip <- sum(u_a * EKrxip_vec)
  T2_diag <- L * C2_vec * (K * Xh^(p + 1) * Rp_ua - dot_a_krxip)

  a_Lp1 <- drop(a_row %*% Lp1)
  u_T3  <- a_Lp1 * as.numeric(t(ep1) %*% iGq)
  u_T3_ELrrq_iGq <- drop(u_T3 %*% ELrrq %*% iGq)
  Rq_uT3         <- as.numeric(Rq %*% u_T3)
  Rq_uT3_ELrrq   <- as.numeric(Rq %*% u_T3_ELrrq_iGq)

  T3_diag <- L * Rq_uT3_ELrrq - L^2 * Rq_uT3 * quadRq

  lbc1_diag <- lus1_diag - rho^(p + 1) * (T1_diag + T2_diag + T3_diag)
  q2 <- sum(lbc1_diag * lbc0 * res^2)

  ## Double-sum terms for (i != j): closed-form k x k contractions.
  ## Equivalent to the original N x N construction but uses only O(N*k + k^2)
  ## memory. Validated to machine precision against the matrix form.
  ##
  ## lbc1_mat[i,j] decomposes (each row vector below is a length-N quantity):
  ##   = alpha[i]                                          # A_mat
  ##   - K[i] * gamma[j] * M_iGp[i,j]                      # B_mat
  ##   - rho^(p+1) * delta[i] * (C1_scalar - epsilon[j])   # T1_mat
  ##   - rho^(p+1) * delta[i] * (Xhpp1[i]*gamma[j] - dot_a_krxp) # T2_mat
  ##   - rho^(p+1) * (zeta[i] - L[i]*eta[j]*M_iGq[i,j])    # T3_mat
  ## with M_iGp = Rp iGp Rp', M_iGq = Rq iGq Rq'.
  dot_a_krxp <- sum(u_a * EKrxp_vec)
  Xhpp1      <- Xh^(p + 1)
  ressq      <- vx                       # res^2

  alpha   <- K * Rp_va
  gamma   <- K * Rp_ua
  delta   <- L * C2_vec
  epsilon <- K * Rp_ua * D2_vec
  zeta    <- L * Rq_uT3_ELrrq
  eta_v   <- L * Rq_uT3

  ## ---- q6: sum_{i!=j} lbc0[i]^2 * L[j]^2 * res[j]^2 * M_iGq[i,j]^2 ----
  ## Identity: sum_{i,j} a[i] b[j] (Rq[i,] iGq Rq[j,]')^2
  ##         = trace( A_a iGq A_b iGq ),
  ##   where A_w = Rq^T diag(w) Rq.
  a_q6  <- lbc0^2
  b_q6  <- (L^2) * ressq
  A_q6  <- crossprod(Rq, a_q6 * Rq)
  B_q6  <- crossprod(Rq, b_q6 * Rq)
  total_q6 <- sum(A_q6 * (iGq %*% B_q6 %*% iGq))
  diag_q6  <- sum(a_q6 * b_q6 * quadRq^2)
  q6 <- total_q6 - diag_q6

  ## ---- q10, q11: sum_{i!=j} lbc1_mat[i,j] * f[i] * g[j] ----
  .lbc1_offdiag <- function(f, g) {
    sum_g   <- sum(g)
    sum_df  <- sum(delta * f)
    fg      <- f * g
    sum_dfg <- sum(delta * fg)

    # A_mat
    St_A <- sum(alpha * f) * sum_g
    Sd_A <- sum(alpha * fg)
    # B_mat
    uB   <- crossprod(Rp, K * f)
    vB   <- crossprod(Rp, gamma * g)
    St_B <- drop(t(uB) %*% iGp %*% vB)
    Sd_B <- sum(K * gamma * quadRp * fg)
    # T1_mat
    St_T1 <- C1_scalar * sum_df * sum_g - sum_df * sum(epsilon * g)
    Sd_T1 <- sum(delta * fg * (C1_scalar - epsilon))
    # T2_mat
    St_T2 <- sum(delta * Xhpp1 * f) * sum(gamma * g) - dot_a_krxp * sum_df * sum_g
    Sd_T2 <- sum(delta * Xhpp1 * gamma * fg) - dot_a_krxp * sum_dfg
    # T3_mat
    uT3   <- crossprod(Rq, L * f)
    vT3   <- crossprod(Rq, eta_v * g)
    St_T3 <- sum(zeta * f) * sum_g - drop(t(uT3) %*% iGq %*% vT3)
    Sd_T3 <- sum(zeta * fg) - sum(L * eta_v * quadRq * fg)

    St <- St_A - St_B - rho^(p + 1) * (St_T1 + St_T2 + St_T3)
    Sd <- Sd_A - Sd_B - rho^(p + 1) * (Sd_T1 + Sd_T2 + Sd_T3)
    St - Sd
  }

  q10 <- .lbc1_offdiag(lbc0 * ressq, lbc0^2 * ressq)
  q11 <- .lbc1_offdiag(lbc0 * ressq, lbc0^2 * ressq - h * s2)

  ## Expectations
  Eq1  <- (q1 / (N * h))^2
  Eq2  <- q2 / (N * h)
  Eq3  <- q3 / (N * h)
  Eq4  <- q4 / (N * h)
  Eq5  <- drop(q5a %*% q5b) / (N * h)^2
  Eq6  <- q6 / (N * (N - 1) * h^2)
  Eq7  <- drop(q7a %*% q7b %*% q7c) / (N * h)^3
  Eq8  <- q8 / (N * h)
  Eq9  <- q9 / (N * h)
  Eq10 <- q10 / (N * (N - 1) * h^2)
  Eq11 <- q11 / (N * (N - 1) * h^2)
  Eq12 <- q12 / (N * h)

  z  <- 1.959964
  pz <- 0.05844507

  q1bc <- pz * (
      Eq1  * (z^3 / 3 + 7 * z / 4 + s2 * z * (z^2 - 3) / 4) / s2^3
    + Eq2  * (-z * (z^2 - 3) / 2) / s2
    + Eq3  * (z * (z^2 - 3) / 8)  / s2^2
    - Eq4  * (z * (z^2 - 1) / 2)  / s2
    - Eq5  * (z * (z^2 - 1))      / s2^2
    + Eq6  * (z * (z^2 - 1) / 4)  / s2
    + Eq7  * (z * (z^2 - 1) / 2)  / s2^2
    + Eq8  * (-z * (z^2 - 3) / 24)/ s2^2
    + Eq9  * (z * (z^2 - 1) / 4)  / s2^2
    + Eq10 * (z * (z^2 - 3))      / s2^2
    + Eq11 * (-z)                 / s2^2
    + Eq12 * (-z * (z^2 + 1) / 8) / s2^2
  )
  q2bc <- -pz * z / (2 * s2)
  Eq3a <- q3a / (N * h)
  q3bc <- pz * Eq3a / s2^2 * (z^3 / 3)

  list(q1rbc = as.numeric(2 * q1bc / pz),
       q2rbc = as.numeric(2 * q2bc / pz),
       q3rbc = as.numeric(2 * q3bc / pz))
}


W.fun = function(u,kernel){
  if (kernel=="epa") w = 0.75*(1-u^2)*(abs(u)<=1)
  if (kernel=="uni") w =          0.5*(abs(u)<=1)
  if (kernel=="tri") w =   (1-abs(u))*(abs(u)<=1)
  if (kernel=="gau") w =   dnorm(u)
  return(w)
}

kd.bw.fun = function(V, B, N, v, r) ((1+2*r)*V/(2*v*N*B^2))^(1/(1+2*v+2*r))

kd.K.fun = function(x,v,r,kernel){
  if (v==2) {
    if (kernel=="gau"){
      if (r==0) k = function(u)            dnorm(u)
      if (r==2) k = function(u)     (u^2-1)*dnorm(u)
      if (r==4) k = function(u) (u^4-6*u^2+3)*dnorm(u)
    }
    if (kernel=="uni"){
      if (r==0) k = function(u)  0.5*(abs(u)<=1)
    }
    if (kernel=="epa"){
      if (r==0) k = function(u)  0.75*(1-u^2)*(abs(u)<=1)
    }
  }
  if (v==4) {
    if (kernel=="uni"){
      if (r==0) k = function(u)    (abs(u)<=1)* 3*(-5*u^2+3)/8
      if (r==2) k = function(u)    (abs(u)<=1)*15*(3*u^2-1)/4
    }
    if (kernel=="epa"){
      if (r==0) k = function(u)   (abs(u)<=1)*(15/32)*(7*u^4-10*u^2+3)
      if (r==2) k = function(u)   (abs(u)<=1)*(105/16)*(6*u^2-5*u^4-1)
    }
  }
  if (v==6) {
    if (kernel=="uni"){
      if (r==0) k = function(u)    (abs(u)<=1)*15*(63*u^4-70*u^2+15)/128
      if (r==2) k = function(u)    (abs(u)<=1)*105*(-45*u^4+42*u^2-5)/32
    }
    if (kernel=="epa"){
      if (r==0) k = function(u)   (abs(u)<=1)*(35/256)*(-99*u^6+189*u^4-105*u^2+15)
      if (r==2) k = function(u)   (abs(u)<=1)*(315/64)*(77*u^6-135*u^4+63*u^2-5)
    }
  }
  Kx = k(x)
  k.v = integrate(function(u) (-1)^v*u^(v)*k(u)/factorial(v), -Inf,Inf)$value
  R.v = integrate(function(u) (k(u))^2,   -Inf,Inf)$value
  out=list(Kx=Kx, k.v=k.v, R.v=R.v)
  return(out)
}

kd.cer.fun = function(x,x0,h,b,v,kernel) {
  n = length(x)
  rho = h/b
  range = max(x)-min(x)

  q.rot = sd(x)*n^(-1/(1+2*v+2*(v+2)))
  K.q = kd.K.fun((x-x0)/q.rot, v=2, r=v+2,   kernel="gau")
  f.r.2 = mean(K.q$Kx)/q.rot^(1+2*v)
  #f.q.rot = kd.K.fun(eval[i], v=2, r=4, kernel="gau")$Kx
  #f.r.2 =  mean((u^4 -6*u^2 + 3)*dnorm(u))/h^(v+3)

  #u = (x-x0)/h
  v.K = kd.K.fun(1, v=v, r=0, kernel=kernel)$k.v
  M.fun = function(u) kd.K.fun(u, v=v,   r=0, kernel=kernel)$Kx - rho^(1+v)*kd.K.fun(rho*u, v=v+2, r=v, kernel=kernel)$Kx*v.K
  K.fun = function(u) kd.K.fun(u, v=v,   r=0, kernel=kernel)$Kx
  L.fun = function(u) kd.K.fun(u, v=v+2, r=v, kernel=kernel)$Kx
  v.fun = function(p,kern) integrate(function(u)   (kern(u))^p,-Inf,Inf)$value
  #m.fun = function(m,kern) ((-1)^m/factorial(m))*(integrate(function(u) u^(m)*kern(u),-Inf,Inf)$value)
  m.fun = function(m,kern) integrate(function(u) u^(m)*kern(u),-Inf,Inf)$value
  v.M.2 = v.fun(2,M.fun)
  v.M.3 = v.fun(3,M.fun)
  v.M.4 = v.fun(4,M.fun)
  m.M.4 = m.fun(4,M.fun)
  m.K.4 = m.fun(4,K.fun)
  m.K.2 = m.fun(2,K.fun)
  m.L.2 = m.fun(2,L.fun)
  
  z = qnorm(0.975)
  q1 = v.M.4*(z^2-3)/6 - v.M.3^2*(z^4-4*z^2+15)/9
  q2 = f.r.2^2 * (m.K.4-rho^(-2)*m.K.2*m.L.2/12)^2 * v.M.2
  q3 = f.r.2   * (m.K.4-rho^(-2)*m.K.2*m.L.2/12)   * v.M.3*(2*z^2)/3

  h <- optimize(function(H) {(H^(-1)*q1 - H^(1+2*(v+2))*q2 + H^(v+2)*q3)^2} , interval=c(.Machine$double.eps, range))$minimum*n^(-1/(v+3))
  return(h)
}


qrXXinv = function(x, ...) {
  #tcrossprod(solve(qr.R(qr(x, tol = 1e-10)), tol = 1e-10))
  #tcrossprod(solve(qr.R(qr(x))))
  chol2inv(chol(crossprod(x)))
}


lp.bw.fun <- function(V, Bsq, p, v, N, kernel) {
  m1 = function(i,j,k)   integrate(function(x) x^i*x^j*k(x),0,Inf)$value
  m2 = function(i,j,k)   integrate(function(x) x^i*x^j*(k(x))^2,0,Inf)$value
  GAMMA = function(p,k) {out=matrix(NA,p+1,p+1); for (i in 0:p) for (j in 0:p) out[i+1,j+1]=m1(i,j,k); return(out)}
  NU    = function(p,k) {out=matrix(NA,p+1,1); for (i in 0:p) out[i+1,1]=m1(i,p+1,k); return(out)}
  PSI   = function(p,k) {out=matrix(NA,p+1,p+1); for (i in 0:p) for (j in 0:p) out[i+1,j+1]=m2(i,j,k); return(out)}
  B.lp  = function(p,k) {out=solve(GAMMA(p,k))%*%NU(p,k); out[1]}
  
  C1.fun = function(p0,v,K) {
    S.inv = solve(GAMMA(p0,K))
    C1 = (S.inv%*%NU(p0,K))[v+1]
    return(C1)
  }
  
  C2.fun = function(p0,v,K) {
    S.inv = solve(GAMMA(p0,K))
    C2 = (S.inv%*%PSI(p0,K)%*%S.inv)[v+1,v+1]
    return(C2)
  }
  
  k.fun = function(u){
    if (kernel=="epa") w = 0.75*(1-u^2)*(abs(u)<=1)
    if (kernel=="uni") w =          0.5*(abs(u)<=1)
    if (kernel=="tri") w =   (1-abs(u))*(abs(u)<=1)
    if (kernel=="gau") w =   dnorm(u)
    return(w)
  }
  
  C1.h <- C1.fun(p0=p, v=v, K=k.fun)
  C2.h <- C2.fun(p0=p, v=v, K=k.fun)
  bw   <- ( ((2*v+1)*C2.h*V) / (2*(p+1-v)*C1.h^2*Bsq*N) )^(1/(2*p+3))
  out=list(bw=bw, C1=C1.h, C2=C2.h)
  return(out)
}

lprobust.res = function(X, y, m, hii, vce, matches, dups, dupsid, d) {
  n = length(y)
  res = matrix(NA,n,1)
  if (vce=="nn") {
    for (pos in 1:n) {
      rpos = dups[pos] - dupsid[pos]
      lpos = dupsid[pos] - 1
      while (lpos+rpos < min(c(matches,n-1))) {
        if (pos-lpos-1 <= 0) rpos = rpos + dups[pos+rpos+1]
        else if (pos+rpos+1>n) lpos = lpos + dups[pos-lpos-1]
        else if ((X[pos]-X[pos-lpos-1]) > (X[pos+rpos+1]-X[pos])) rpos = rpos + dups[pos+rpos+1]
        else if ((X[pos]-X[pos-lpos-1]) < (X[pos+rpos+1]-X[pos])) lpos = lpos + dups[pos-lpos-1]
        else {
          rpos = rpos + dups[pos+rpos+1]
          lpos = lpos + dups[pos-lpos-1]
        }
      }
      ind.J = (pos-lpos):min(c(n,(pos+rpos)))
      y.J   = sum(y[ind.J])-y[pos]
      Ji = length(ind.J)-1
      res[pos,1] = sqrt(Ji/(Ji+1))*(y[pos] - y.J/Ji)
    }
  }
  else {
    if (vce=="hc0") w = 1
    else if (vce=="hc1") w = sqrt(n/(n-d))
    else if (vce=="hc2") w = sqrt(1/(1-hii))
    else                 w =      1/(1-hii)
    res[,1] = w*(y-m[,1])
  }
  return(res)
}


## Cluster-robust meat for CR0, CR1, CR2, CR3.
## X:       N x k design (standardized -- sqrt(W)-scaled -- for the classical
##          variance; Q for the bias-corrected variance).
## r:       N-long residuals (standardized sqrt(W) * raw for classical; raw
##          for bias-corrected).
## C:       cluster IDs.
## invG:    k x k inverse Gram matching the design choice.
## cr_type: one of "CR0", "CR1", "CR2", "CR3".
lprobust.cluster.meat = function(X, r, C, invG, cr_type, k_override = NULL) {
  # k_override: when supplied, overrides the (N-1)/(N-k) df correction's k
  # for CR1. Used when X is a "score-like" matrix (e.g. Q.q) whose ncol is
  # smaller than the effective number of regressors that produced r. Without
  # it, the bias-corrected CR1 path used k=p+1 from ncol(Q.q), inconsistent
  # with the q-regression that produced the residuals (k = q+1).
  # Note: k is the matrix dimension; k_df is what enters the CR1 mult.
  clusters <- unique(C)
  G <- length(clusters)
  N <- length(r)
  k    <- ncol(X)
  k_df <- if (is.null(k_override)) k else k_override
  M <- matrix(0, k, k)

  for (cl in clusters) {
    ind <- C == cl
    Xg  <- X[ind, , drop = FALSE]
    rg  <- as.numeric(r[ind])

    if (cr_type %in% c("CR0", "CR1")) {
      r_adj <- rg
    } else {
      H_gg <- Xg %*% invG %*% t(Xg)
      H_gg <- (H_gg + t(H_gg)) / 2          # enforce symmetry numerically
      I_H  <- diag(nrow(Xg)) - H_gg
      if (cr_type == "CR2") {
        ev <- eigen(I_H, symmetric = TRUE)
        vals <- pmax(ev$values, 1e-12)
        r_adj <- as.numeric(ev$vectors %*% (t(ev$vectors) %*% rg / sqrt(vals)))
      } else {   # CR3
        r_adj <- tryCatch(drop(solve(I_H, rg)),
                          error = function(e) {
                            ev <- eigen(I_H, symmetric = TRUE)
                            vals <- pmax(ev$values, 1e-12)
                            as.numeric(ev$vectors %*% (t(ev$vectors) %*% rg / vals))
                          })
      }
    }
    score <- crossprod(Xg, r_adj)
    M <- M + tcrossprod(score)
  }

  mult <- switch(cr_type,
                 "CR0" = 1,
                 "CR1" = ((N - 1) / (N - k_df)) * (G / (G - 1)),
                 "CR2" = 1,
                 "CR3" = (G - 1) / G)
  mult * M
}

lprobust.vce = function(RX, res, C) {
  n = length(C)
  k = ncol(RX)
  M = matrix(0, k, k)
  if (is.null(C)) {
    M = crossprod(c(res) * RX)
  } else {
    clusters = unique(C)
    g        = length(clusters)
    w        = ((n - 1) / (n - k)) * (g / (g - 1))
    for (i in 1:g) {
      ind = C == clusters[i]
      Xi  = RX[ind, , drop = FALSE]
      ri  = res[ind, , drop = FALSE]
      score = crossprod(Xi, ri)
      M = M + tcrossprod(score)
    }
    M = w * M
  }
  return(M)
}

lprobust.bw = function(Y, X, cluster, c, o, nu, o.B, h.V, h.B1, h.B2, scale, vce, nnmatch, kernel, dups, dupsid, weights = NULL) {

  if (is.null(weights)) weights <- rep(1, length(X))

  ### Variance
  dC = 0
  eC = NULL
  w = W.fun((X-c)/h.V, kernel)/h.V * weights
  ind.V = w> 0; eY = Y[ind.V];eX = X[ind.V];eW = w[ind.V]
  n.V = sum(ind.V)
  R.V = matrix(NA,n.V,o+1)
  for (j in 1:(o+1)) R.V[,j] = (eX-c)^(j-1)
  invG.V = qrXXinv(R.V*sqrt(eW))
  beta.V = invG.V%*%crossprod(R.V*eW,eY)
  dups.V = dupsid.V = predicts.V = 0

  if (!is.null(cluster)) {
    dC = 1
    eC =  cluster[ind.V] 
  }
  
  if (vce=="nn") {
    dups.V   = dups[ind.V]
    dupsid.V = dupsid[ind.V]
  }

  if (vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") {
    predicts.V=R.V%*%beta.V
    if (vce=="hc2" | vce=="hc3") {
      hii <- matrix(rowSums((R.V %*% invG.V) * (R.V * eW)), n.V, 1)
    }
  }
  res.V = lprobust.res(eX, eY, predicts.V, hii, vce, nnmatch, dups.V, dupsid.V, o+1)
  V.V   = (invG.V%*%lprobust.vce(R.V*eW, res.V, eC)%*%invG.V)[nu+1,nu+1]
  
  ### BIAS
  #Hp = diag(c(1,poly(h.V,degree=o,raw=TRUE)))
  Hp = 0
  for (j in 1:(o+1)) Hp[j] = h.V^((j-1))
  v1 = crossprod(R.V*eW,((eX-c)/h.V)^(o+1))
  v2 = crossprod(R.V*eW,((eX-c)/h.V)^(o+2))
  BConst1 = (Hp*(invG.V%*%v1))[nu+1]
  BConst2 = (Hp*(invG.V%*%v2))[nu+1]

  w = W.fun((X-c)/h.B1, kernel) * weights
  ind = w> 0
  n.B = sum(ind)
  eY = Y[ind];eX = X[ind];eW = w[ind]

  if (!is.null(cluster)) eC =  cluster[ind]

  R.B1 = matrix(NA,n.B,o.B+1)
  for (j in 1:(o.B+1)) R.B1[,j] = (eX-c)^(j-1)
  invG.B1 = qrXXinv(R.B1*sqrt(eW))
  beta.B1 = invG.B1%*%crossprod(R.B1*eW,eY)

  BWreg=0
  if (scale>0) {
    #e.B = matrix(0,(o.B+1),1); e.B[o+2]=1
    dups.B = dupsid.B = hii = predicts.B = 0
    if (vce=="nn") {
      dups.B   = dups[ind]
      dupsid.B = dupsid[ind]
    }
    if (vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") {
      predicts.B = R.B1%*%beta.B1
      if (vce=="hc2" | vce=="hc3") {
        hii <- matrix(rowSums((R.B1 %*% invG.B1) * (R.B1 * eW)), n.B, 1)
      }
    }
    res.B = lprobust.res(eX, eY, predicts.B, hii, vce, nnmatch, dups.B, dupsid.B,o.B+1)
    V.B = (invG.B1%*%lprobust.vce(R.B1*eW, res.B, eC)%*%invG.B1)[o+2,o+2]
    BWreg = 3*BConst1^2*V.B
  }
  
  w = W.fun((X-c)/h.B2, kernel) * weights
  ind = w> 0
  n.B = sum(ind)
  eY = Y[ind];eX = X[ind];eW = w[ind]

  R.B2 = matrix(NA,n.B,o.B+2)
  for (j in 1:(o.B+2)) R.B2[,j] = (eX-c)^(j-1)
  invG.B2 = qrXXinv(R.B2*sqrt(eW))
  beta.B2 = invG.B2%*%crossprod(R.B2*eW,eY)
  
  N  <- length(X)
  B1 <- BConst1%*%(beta.B1[o+2,])
  B2 <- BConst2%*%(beta.B2[o+3,])
  V  <- N*h.V^(2*nu+1)*V.V
  R  <- BWreg
  r  <- 1/(2*o+3)
  rB <- 2*(o+1-nu)
  rV <- 2*nu+1
  bw <- ( (rV*V) / (N*rB*(B1^2 + scale*R)) )^r
  
  output = list(V=V, B1=B1, B2=B2, R=R, r=r, rB=rB, rV=rV, bw=bw)
  return(output)
}


lpbwselect.ce.dpi = function(y, x, h, b, eval, p, q, deriv, rho, kernel, vce, nnmatch, interior, bwregul, weights = NULL){

  if (is.null(weights)) weights <- rep(1, length(x))
  
  rho <- 1
  bwregul <- 0
  
  N = length(x)
  range = max(x)-min(x)
  
  dups <- dupsid <- hii <- predicts <- NULL


  if (vce=="nn") {
    order.x <- order(x)
    x <- x[order.x]
    y <- y[order.x]
  }
  
  if (!is.null(rho)){
    b <- h/rho
  } else {
    rho <- h/b
  }

  X.h <- (x-eval)/h;  X.b <- (x-eval)/b
  K.h <- W.fun(X.h,kernel) * weights
  L.b <- W.fun(X.b,kernel) * weights
  ind.h <- K.h>0;    ind.b <- L.b>0
  ind <- ind.h | ind.b
  eN   <- sum(ind)
  eY   <-   y[ind];  eX   <-   x[ind]
  eX.h <- X.h[ind];  eX.b <- X.b[ind]
  eK.h <- K.h[ind];  eL.b <- L.b[ind]

  W.p <- eK.h/h;  W.q <- eL.b/b
  R.p.2 <- matrix(NA,eN,(p+3))
  for (j in 1:(p+3))  R.p.2[,j] <- eX.h^(j-1)
  R.p.1 <- R.p.2[,1:(p+2)]
  R.p   <- R.p.2[,1:(p+1)]
  R.q   <- matrix(NA,eN,(q+1))
  for (j in 1:(q+1))  R.q[,j] <- eX.b^(j-1)

  L.p.1 <- crossprod(R.p*W.p, eX.h^(p+1))/eN
  L.p.2 <- crossprod(R.p*W.p, eX.h^(p+2))/eN
  L.p.3 <- crossprod(R.p*W.p, eX.h^(p+3))/eN
  L.q.1 <- crossprod(R.q*W.q, eX.b^(q+1))/eN
  L.q.2 <- crossprod(R.q*W.q, eX.b^(q+2))/eN
  L.q.3 <- crossprod(R.q*W.q, eX.b^(q+3))/eN
  invG.p <- eN*qrXXinv(R.p*sqrt(W.p))
  invG.q <- eN*qrXXinv(R.q*sqrt(W.q))

  edups <- edupsid <- 0
  if (vce=="nn") {
    runs    <- rle(eX)
    edups   <- rep.int(runs$lengths, runs$lengths)
    edupsid <- sequence(runs$lengths)
  }
  
  hii=predicts.p=predicts.q=0
  if (vce=="hc0" | vce=="hc1" | vce=="hc2" | vce=="hc3") {
    H.q <- 0
    for (j in 1:(q+1)) H.q[j] <- b^(-(j-1))
    beta.q <- H.q*invG.q%*%crossprod(R.q*W.q,eY)/eN
    r.q <- matrix(NA,eN,q+1)
    predicts <- 0
    for (j in 1:(q+1))  r.q[,j] <- (eX-eval)^(j-1)
    for (j in 1:eN) predicts[j] <- r.q[j,]%*%beta.q
    if (vce=="hc2" | vce=="hc3") {
      hii <- matrix(rowSums((R.p %*% invG.p) * (R.p * W.p)) / eN, eN, 1)
    }
  }

  res.q <- lprobust.res(eX, eY, as.matrix(predicts), hii, vce, nnmatch, edups, edupsid, q+1)
 
  ### Bias
  k <- p+3
  r.k <- matrix(NA,N,k+3)
  for (j in 1:(k+3))  r.k[,j] <- x^(j-1)
  gamma <- lm(y~r.k-1)
  m.p.3 <- gamma$coeff[p+4]*factorial(p+3) + gamma$coeff[p+5]*factorial(p+4)*eval + gamma$coeff[p+6]*factorial(p+5)*eval^2/2
  #r.k <- r.k[,1:(k+2)]
  gamma <- lm(y~r.k[,1:(k+2)]-1)
  m.p.2 <- gamma$coeff[p+3]*factorial(p+2) + gamma$coeff[p+4]*factorial(p+3)*eval + gamma$coeff[p+5]*factorial(p+4)*eval^2/2

  if (is.na(m.p.3)) m.p.3 <- lprobust(y=y, x=x, h=range, eval=eval, p=p+4, deriv=(p+3), kernel=kernel, vce=vce)$Estimate[5]
  if (is.na(m.p.2)) m.p.2 <- lprobust(y=y, x=x, h=range, eval=eval, p=p+4, deriv=(p+2), kernel=kernel, vce=vce)$Estimate[5]
  
  e.p.1 <- matrix(0,(q+1),1); e.p.1[p+2]=1
  e.0   <- matrix(0,(p+1),1); e.0[1]=1

  q.terms <- lpbwce(y = eY, x = eX,  K=eK.h, L=eL.b, res=res.q, c = eval, p=p, q=q, h=h, b=b, deriv=deriv, fact=factorial(deriv))
  q1.rbc  <- q.terms$q1rbc
  q2.rbc  <- q.terms$q2rbc
  q3.rbc  <- q.terms$q3rbc

  V.reg=0
  if (bwregul>0) {
    o <- p+2
    R.reg <-  matrix(NA,eN,o+1)
    for (j in 1:(o+1)) R.reg[,j] = (eX-eval)^(j-1)
    invG.reg = qrXXinv(R.reg*sqrt(eK.h))
    V.reg = (invG.reg%*%lprobust.vce(R.reg*eK.h, res.q)%*%invG.reg)[o+1,o+1]
    #beta.reg = invG.reg%*%crossprod(R.reg*eK.h,eY)
    #Hp = 0
    #for (j in 1:o) Hp[j] = h^((j-1))
    #v = crossprod(R.reg*eK.h,((eX-eval)/h)^o)
    #BConst = (Hp*(invG.reg%*%v))[1]
  }  
  
  if (isTRUE(interior)) {
    eta.bc1  <- (t(e.0)%*%invG.p)%*%(   (m.p.2/factorial(p+2))*L.p.2/h     + (m.p.3/factorial(p+3))*L.p.3 )
    eta.bc2  <- rho^(-2)*b^(q-p-1)*(t(e.0)%*%invG.p)%*%L.p.1%*%t(e.p.1)%*%invG.q%*%( (m.p.2/factorial(p+2))*L.q.1/b     + (m.p.3/factorial(p+3))*L.q.2 )
    eta.bc   <- (eta.bc1 - eta.bc2)
    Reg      <- 3*(t(e.0)%*%invG.p)%*%(L.p.2/h)^2*V.reg
    
    H.bc     <- function(H) {abs(H^(-1)*q1.rbc + H^(1+2*(p+3))*(eta.bc^2 + bwregul*Reg)*q2.rbc + H^(p+3)*(eta.bc + bwregul*Reg)*q3.rbc)}
    h.bc     <- optimize(H.bc , interval=c(.Machine$double.eps, range))
    h.ce.dpi <- h.bc$minimum*N^(-1/(p+4))
  }

  if (isFALSE(interior)) {
    eta.bc.1 <- (t(e.0)%*%invG.p)%*%( L.p.2 - rho^(-1)*L.p.1%*%t(e.p.1)%*%invG.q%*%L.q.1)*(m.p.2/factorial(p+2))
    eta.bc.2 <- (t(e.0)%*%invG.p)%*%( L.p.3 - rho^(-2)*L.p.1%*%t(e.p.1)%*%invG.q%*%L.q.2)*(m.p.3/factorial(p+3))
    Reg      <- 3*((t(e.0)%*%invG.p)%*%( L.p.2 - rho^(-1)*L.p.1%*%t(e.p.1)%*%invG.q%*%L.q.1))^2*V.reg
    
    phi.z    <- dnorm(1.96)
    E1       <- q1.rbc*phi.z
    E2       <- eta.bc.1
    E3       <- eta.bc.2
    E4       <- q3.rbc*phi.z*eta.bc.1
    E5       <- q3.rbc*phi.z*eta.bc.2
    H.bc <-function(H) {abs(E1/(N*H) + N*H^(2*p+5)*q2.rbc*phi.z*(E2 + H*E3 + bwregul*Reg)^2 + H^(p+2)*(E4 + H*E5 + bwregul*Reg))}
    h.ce.dpi <- optimize(H.bc , interval=c(.Machine$double.eps, range))$minimum
  }
  out <- list(h=h.ce.dpi)
  return(out)
}

lpbwselect.imse.dpi = function(y, x, cluster, p, q, deriv, kernel, bwcheck, bwregul, imsegrid, vce, nnmatch, interior, weights = NULL){

  N <- length(x)
  x.max <- max(x);  x.min <- min(x)
  range <- x.max - x.min
  eval <- seq(x.min, x.max, length.out=imsegrid)
  neval = length(eval)
  even <- (p-deriv)%%2==0
  V.h <- B.h <- V.b <- B.b <- 0

  for (i in 1:neval) {
    est <- lpbwselect.mse.dpi(y=y, x=x, cluster=cluster, weights=weights,
                              eval=eval[i], p=p, q=q, deriv=deriv,
                              kernel=kernel, bwcheck=bwcheck, bwregul=bwregul,
                              vce=vce, nnmatch=nnmatch, interior=interior)

    V.h[i] <- est$V.h;    B.h[i] <- est$B.h
    V.b[i] <- est$V.b;    B.b[i] <- est$B.b
  }

  if (isFALSE(even) | isTRUE(interior)) {
    b.imse.dpi <- (mean(V.b)/(N*mean(B.b)))^(1/(2*q+3))
    h.imse.dpi <- (mean(V.h)/(N*mean(B.h)))^(1/(2*p+3))
  }

  if (isTRUE(even) & isFALSE(interior)) {
    b.bw.fun   <- function(H) {abs(H^(2*q+2-2*(p+1))*mean(B.b) + mean(V.b)/(N*H^(1+2*(p+1))))}
    b.imse.dpi <- optimize(b.bw.fun, interval=c(.Machine$double.eps, range))$minimum

    h.bw.fun   <- function(H) {abs(H^(2*p+2-2*deriv)*mean(B.h) + mean(V.h)/(N*H^(1+2*deriv)))}
    h.imse.dpi <- optimize(h.bw.fun, interval=c(.Machine$double.eps, range))$minimum
  }

  out <- list(h=h.imse.dpi, b=b.imse.dpi)
  return(out)
}


lpbwselect.imse.rot = function(y, x, p, deriv, kernel, imsegrid){
  
  even <- (p-deriv)%%2==0
  eval <- quantile(x, probs = seq(0.05, 0.95, 0.025))
  #eval  <- seq(min(x), max(x), length.out=imsegrid)
  neval <- length(eval)
  x.max <- max(x);  x.min <- min(x)
  range <- x.max - x.min
  N     <- length(x)
  V     <- B <- 0

  for (i in 1:neval) {
    est  <- lpbwselect.mse.rot(y=y, x=x, eval=eval[i], p=p, deriv=deriv, kernel=kernel)
    V[i] <- est$V;    B[i] <- est$B
  }
  
  if (isFALSE(even)) {
    h.imse.rot <- (mean((1+2*deriv)*V)/(N*mean(2*(p+1-deriv)*B^2)))^(1/(2*p+3))
  } else {
    h.bw.fun   <- function(H) {abs(H^(2*p+2-2*deriv)*mean(B^2) + mean(V)/(N*H^(1+2*deriv)))}
    h.imse.rot <- optimize(h.bw.fun, interval=c(.Machine$double.eps, range))$minimum
  }
  out <- list(h=h.imse.rot)
  return(out)
}


lpbwselect.mse.dpi = function(y, x, cluster, eval, p, q, deriv, kernel, bwcheck, bwregul, vce, nnmatch, interior, weights = NULL){

  even <- (p-deriv)%%2==0

                     C.c         <- 2.34
  if (kernel=="uni") C.c         <- 1.843
  if (kernel=="tri") C.c         <- 2.576
  if (kernel=="gau") C.c         <- 1.06

  x.iq  <- quantile(x,.75) - quantile(x,.25)
  x.max <- max(x);  x.min <- min(x)
  range <- x.max - x.min
  N     <- length(x)

  bw.max = max(abs(eval-x.min),abs(eval-x.max))
  
  c.bw  <- c(C.c*min(c(sd(x),x.iq/1.349))*N^(-1/5))
  c.bw <- min(c.bw, bw.max)
  
  dups <- dupsid <- hii <- predicts <- NULL
  if (vce=="nn") {
    order.x = order(x)
    x = x[order.x];   y = y[order.x]
    runs   <- rle(x)
    dups   <- rep.int(runs$lengths, runs$lengths)
    dupsid <- sequence(runs$lengths)
  }

  bw.adj <- 0
  if (!is.null(bwcheck)) {
    bw.min    <- sort(abs(x-eval))[bwcheck]
    #nh        <- sum(abs(x-eval) <= c.bw)
    c.bw      <- max(c.bw, bw.min)
    bw.adj <- 1
  }

   C.d1 <- lprobust.bw(y, x, cluster, c=eval, o=q+1, nu=q+1, o.B=q+2, h.V=c.bw, h.B1=range, h.B2=range, 0, vce, nnmatch, kernel, dups, dupsid, weights = weights)
   if (isFALSE(even) | isTRUE(interior)) {
       bw.mp2 <- c(C.d1$bw)
    }
    else{
      bw.fun  <-function(H) {abs(H^(2*(q+1)+2-2*(q+1))*(C.d1$B1 + H*C.d1$B2)^2 + C.d1$V/(N*H^(1+2*(q+1))))}
      bw.mp2 <- optimize(bw.fun, interval=c(.Machine$double.eps, range))$minimum
   }
   

    C.d2 <- lprobust.bw(y, x, cluster, c=eval, o=q+2, nu=q+2, o.B=q+3, h.V=c.bw, h.B1=range, h.B2=range, 0, vce, nnmatch, kernel, dups, dupsid, weights = weights)
    if (isFALSE(even) | isTRUE(interior)) {
      bw.mp3 <- c(C.d2$bw)
    }
    else {
      bw.fun  <-function(H) {abs(H^(2*(q+2)+2-2*(q+2))*(C.d2$B1 + H*C.d2$B2)^2 + C.d2$V/(N*H^(1+2*(q+2))))}
      bw.mp3 <- optimize(bw.fun, interval=c(.Machine$double.eps, range))$minimum
    }
    
    # adjust
    bw.mp2 <- min(bw.mp2, bw.max) 
    bw.mp3 <- min(bw.mp3, bw.max) 
    
    if (!is.null(bwcheck)) {
      bw.mp2 <- max(bw.mp2, bw.min)
      bw.mp3 <- max(bw.mp3, bw.min)
    }
    
    
    ### Select preliminar bw b
    C.b <- lprobust.bw(y, x, cluster, c=eval, o=q, nu=p+1, o.B=q+1, h.V=c.bw, h.B1=bw.mp2, h.B2=bw.mp3, bwregul, vce, nnmatch, kernel, dups, dupsid, weights = weights)
    
    if (isFALSE(even) | isTRUE(interior)) {
      b.mse.dpi <- c(C.b$bw)
    }
    else {
      b.bw.fun = function(H) {abs(H^(2*q+2-2*(p+1))*(C.b$B1 + H*C.b$B2 + bwregul*C.b$R)^2 + C.b$V/(N*H^(1+2*(p+1))))}
      b.mse.dpi <- optimize(b.bw.fun, interval=c(.Machine$double.eps, range))$minimum
    }
    
    b.mse.dpi <- min(b.mse.dpi, bw.max) 
    if (!is.null(bwcheck)) b.mse.dpi <- max(b.mse.dpi, bw.min)
  
    bw.mp1 = b.mse.dpi
    
    ### Select final bw h
    C.h <- lprobust.bw(y, x, cluster, c=eval, o=p, nu=deriv, o.B=q, h.V=c.bw, h.B1=bw.mp1, h.B2=bw.mp2, bwregul, vce, nnmatch, kernel, dups, dupsid, weights = weights)
    
    if (isFALSE(even) | isTRUE(interior)) {
      h.mse.dpi <- c(C.h$bw)
    } else {
      h.bw.fun = function(H) {abs(H^(2*p+2-2*deriv)*(C.h$B1 + H*C.h$B2 + bwregul*C.h$R)^2 + C.h$V/(N*H^(1+2*deriv)))}
      h.mse.dpi <- optimize(h.bw.fun, interval=c(.Machine$double.eps, range))$minimum
    }
    
    h.mse.dpi <- min(h.mse.dpi, bw.max) 
    if (!is.null(bwcheck)) h.mse.dpi <- max(h.mse.dpi, bw.min)
  
    if (isFALSE(even) | isTRUE(interior)) {
      V.h <- C.h$rV*C.h$V;    B.h <- C.h$rB*C.h$B1^2
      V.b <- C.b$rV*C.b$V;    B.b <- C.b$rB*C.b$B1^2
    } else {
      V.h <- C.h$V;    B.h <- (C.h$B1 + h.mse.dpi*C.h$B2)^2
      V.b <- C.b$V;    B.b <- (C.b$B1 + b.mse.dpi*C.b$B2)^2
    }

  out <- list(h=h.mse.dpi, b=b.mse.dpi, V.h=V.h, B.h=B.h, V.b=V.b, B.b=B.b)
  return(out)
}

lpbwselect.mse.rot = function(y, x, eval, p, deriv, kernel){

  even <- (p-deriv)%%2==0
  
  C.c         <- 2.34
  if (kernel=="uni") C.c         <- 1.843
  if (kernel=="tri") C.c         <- 2.576

  x.iq  <- quantile(x,.75) - quantile(x,.25)
  x.max <- max(x);  x.min <- min(x)
  range <- x.max - x.min
  N     <- length(x)

  c.bw  <- C.c*min(c(sd(x),x.iq/1.349))*N^(-1/5)
  n.h1  <- sum(abs(x-eval) <= c.bw)
  f.hat <- n.h1/(2*N*c.bw)
  
  k   <- p+3
  r.k <- matrix(NA,N,k+1)
  for (j in 1:(k+1))  r.k[,j] <- x^(j-1)
  gamma.p <- lm(y~r.k-1)
  s2.hat  <- sigma(gamma.p)^2

  #k   <- q+3
  #r.k <- matrix(NA,N,k+1)
  #for (j in 1:(k+1))  r.k[,j] <- x^(j-1)
  #gamma.q <-lm(y~r.k-1)
  #s2.q    <- sigma(gamma.q)^2

  m.p.1 <- lprobust(y=y, x=x, h=range, eval=eval, p=p+3, deriv=(p+1), kernel=kernel, vce="nn")$Estimate[5]
  m.p.2 <- lprobust(y=y, x=x, h=range, eval=eval, p=p+3, deriv=(p+2), kernel=kernel, vce="nn")$Estimate[5]
  
  #m.q.1 <- lprobust(y=y, x=x, h=range, eval=eval, p=q+3, deriv=(q+1), kernel=kernel, vce=vce)$Estimate[5]
  #m.q.2 <- lprobust(y=y, x=x, h=range, eval=eval, p=q+3, deriv=(q+1), kernel=kernel, vce=vce)$Estimate[5]
  
  #2*gamma.p$coeff[p+2] + 3*2*gamma.p$coeff[p+3]*eval + 4*3*gamma.p$coeff[p+4]*eval^2
  #m.p.1     <- gamma.p$coeff[p+1]*factorial(p+1) + gamma.p$coeff[p+2]*factorial(p+2)*eval + gamma.p$coeff[p+3]*factorial(p+3)*eval^2/2
  #m.q.1     <- gamma.q$coeff[q+1]*factorial(q+1) + gamma.q$coeff[q+2]*factorial(q+2)*eval + gamma.q$coeff[q+3]*factorial(q+3)*eval^2/2


  #h.mse.rot <- lp.bw.fun(s2.p/f0.pilot, (m.p.1/factorial(p+1))^2, p, deriv, N, kernel)
  #b.mse.rot <- lp.bw.fun(s2.q/f0.pilot, (m.q.1/factorial(q+1))^2, q, p+1  , N, kernel)

  bw <- lp.bw.fun(s2.hat/f.hat, (m.p.1/factorial(p+1))^2, p, deriv, N, kernel)
  
  B1 = bw$C1*m.p.1/factorial(p+1)
  B2 = bw$C1*m.p.2/factorial(p+2)
  V  = bw$C2*s2.hat/f.hat

  if (isFALSE(even)) {
    h.mse.rot <- bw$bw
    B <- B1
  } else {
    h.bw.fun  <- function(H) {abs(H^(2*p+2-2*deriv)*(B1 + H*B2)^2 + V/(N*H^(1+2*deriv)))}
    h.mse.rot <- optimize(h.bw.fun, interval=c(.Machine$double.eps, range))$minimum
    B <- B1 + h.mse.rot*B2
  }
  
  out <- list(h=h.mse.rot, V=V, B=B)
  return(out)
}

## Regression test pinning lpbwce numerical output.
## Reference values were generated from the original O(N^2) matrix
## implementation; the current closed-form contraction must agree to
## machine precision.

test_that("lpbwce closed-form matches O(N^2) reference", {
  set.seed(42)
  n   <- 400
  x   <- sort(runif(n, -1, 1))
  y   <- 1 + 2 * x + 3 * x^2 + rnorm(n, sd = 0.3)
  res <- y - (1 + 2 * x + 3 * x^2)
  h   <- 0.6
  b   <- 0.9
  K   <- nprobust:::W.fun((x - 0) / h, "epa")
  L   <- nprobust:::W.fun((x - 0) / b, "epa")

  out <- nprobust:::lpbwce(y = y, x = x, K = K, L = L, res = res,
                           c = 0, p = 1, q = 2, h = h, b = b,
                           deriv = 0, fact = 1)

  expect_equal(out$q1rbc, -8.71677124715237, tolerance = 1e-12)
  expect_equal(out$q2rbc, -11.0080523759492, tolerance = 1e-12)
  expect_equal(out$q3rbc,   2.85555479540377, tolerance = 1e-12)
})

kdrobust <- function(x, eval=NULL, neval=NULL, h=NULL, b=NULL, rho=1, kernel="epa",
                    bwselect=NULL, bwcheck=21, imsegrid=30, level=95, subset = NULL,
                    data = NULL) {

  if (!is.null(data)) {
    mc <- match.call()
    caller_env <- parent.frame()
    .lookup <- function(arg) {
      expr <- mc[[arg]]
      if (is.null(expr)) return(NULL)
      eval(expr, envir = data, enclos = caller_env)
    }
    x <- .lookup("x")
    if ("subset" %in% names(mc)) subset <- .lookup("subset")
  }

  p <- 2
  deriv <- 0
  if (!is.null(subset)) x <- x[subset]

  ## UX prechecks: catch invalid h/b/bwcheck/imsegrid with clear messages
  .nperrs <- character()
  .bwbad <- function(v, name) {
    if (is.null(v))         return(NULL)
    if (!is.numeric(v))     return(paste0(name, " must be numeric."))
    if (any(!is.finite(v))) return(paste0(name, " contains non-finite values (NA/Inf)."))
    if (any(v <= 0))        return(paste0(name, " must be strictly positive."))
    NULL
  }
  .nperrs <- c(.nperrs, .bwbad(h, "h"), .bwbad(b, "b"))
  if (!is.null(bwcheck) && (!is.numeric(bwcheck) || length(bwcheck) != 1 || !is.finite(bwcheck) || bwcheck <= 0))
    .nperrs <- c(.nperrs, "bwcheck must be a single positive finite integer.")
  if (!is.numeric(imsegrid) || length(imsegrid) != 1 || !is.finite(imsegrid) || imsegrid <= 0)
    .nperrs <- c(.nperrs, "imsegrid must be a single positive finite integer.")
  if (!is.numeric(level) || length(level) != 1 || !is.finite(level) || level <= 0 || level >= 100)
    .nperrs <- c(.nperrs, "level must be a single number in (0, 100).")
  if (!is.numeric(rho) || length(rho) != 1 || !is.finite(rho) || rho < 0)
    .nperrs <- c(.nperrs, "rho must be a single non-negative number.")
  if (!is.null(eval)) {
    if (!is.numeric(eval) || any(!is.finite(eval)))
      .nperrs <- c(.nperrs, "eval must be numeric and finite.")
    if (length(eval) == 0L)
      .nperrs <- c(.nperrs, "eval must have at least one element.")
  }
  if (length(.nperrs) > 0) {
    for (.m in .nperrs) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  na.ok <- complete.cases(x)
  x <- x[na.ok]
  
  x.min <- min(x);  x.max <- max(x)
  N <- length(x)

  if (!is.null(bwcheck)) {
    if (bwcheck > N) {
      warning("bwcheck (", bwcheck, ") is larger than the sample size (", N,
              "); reducing bwcheck to N.")
      bwcheck <- N
    }
  }

  if (is.null(eval)) {
    if (is.null(neval)) {
      qseq <- seq(0.1,0.9,length.out=30)
      eval <- quantile(x, qseq)
      #eval <- unique(x)
      #qseq <- seq(0,1,1/(20+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      #eval <- seq(x.min, x.max, length.out=30)
    }
    else {
      qseq <- seq(0.1,0.9,length.out=neval)
      eval <- quantile(x, qseq)
      #eval <- seq(x.min,x.max,length.out=neval)
      #qseq <- seq(0,1,1/(neval+1))
      #eval <- quantile(x, qseq[2:(length(qseq)-1)])
      #eval <- seq(x.min, x.max, length.out=neval)
    }
  }
  neval <- length(eval)

  ## Precheck (continued): h/b length must be 1 or neval
  .nperrs2 <- character()
  if (!is.null(h) && length(h) != 1L && length(h) != neval)
    .nperrs2 <- c(.nperrs2, paste0("h must have length 1 or neval (=", neval, ")."))
  if (!is.null(b) && length(b) != 1L && length(b) != neval)
    .nperrs2 <- c(.nperrs2, paste0("b must have length 1 or neval (=", neval, ")."))
  if (length(.nperrs2) > 0) {
    for (.m in .nperrs2) warning(.m, call. = FALSE)
    stop("nprobust: invalid input (see warnings above).", call. = FALSE)
  }

  if (is.null(h) & is.null(bwselect) & neval==1) bwselect="mse-dpi"
  if (is.null(h) & is.null(bwselect) & neval>1)  bwselect="imse-dpi"
  
  kernel   <- tolower(kernel)
  bwselect <- tolower(bwselect)

  #####################################################   CHECK ERRORS
  if (!(kernel %in% c("epa","epanechnikov","uni","uniform"))) {
    stop("kernel incorrectly specified. Supported kernels for kdrobust: epa, uni.")
  }
  if (kernel %in% c("epanechnikov")) kernel <- "epa"
  if (kernel %in% c("uniform"))      kernel <- "uni"

  if (identical(bwselect, "all"))
    stop("bwselect=\"all\" is only supported by kdbwselect; kdrobust requires a single method (e.g. \"mse-dpi\", \"imse-dpi\", \"ce-dpi\").",
         call. = FALSE)

  if (!is.null(h)) bwselect <- "Manual"

  kernel.type <- if (kernel == "epa") "Epanechnikov" else "Uniform"

  if (!is.null(h) & rho>0 & is.null(b)) {
    #rho <- rep(1,neval)
    b <- h/rho
  }
  #if (!is.null(h) & !is.null(rho) ) b <- h/rho
  if (is.null(h)) {
      kdbws <- kdbwselect(x=x, eval=eval, bwselect=bwselect, bwcheck=bwcheck, imsegrid=imsegrid, kernel=kernel)
      h <- kdbws$bws[,2]
      b <- kdbws$bws[,3]
      if (rho>0) b <- h/rho
      rho <- h/b  
  }
  
  if (length(h)==1 & neval>1) {
    h <- rep(h,neval)
    b <- rep(b,neval)
    rho   <- h/b
  }

  
  Estimate<-matrix(NA,neval,8)
  colnames(Estimate)<-c("eval","h","b","N","tau.us","tau.bc","se.us","se.rb")
  
  for (i in 1:neval) {
    
    if (!is.null(bwcheck)) {
      bw.min   <- sort(abs(x-eval[i]))[bwcheck]
      h[i]     <- max(h[i], bw.min)
      b[i]     <- max(b[i], bw.min)
      rho[i]   <- h[i]/b[i]
    }
    
    u   <- (x-eval[i])/h[i]
    K.d <- kd.K.fun(u        , v=p,   r=deriv, kernel=kernel)
    L.r <- kd.K.fun(rho[i]*u , v=p+2, r=p,     kernel=kernel)
  
    K     <- K.d$Kx
    M     <- K - rho[i]^(1+p)*L.r$Kx*L.r$k.v
    f.us  <- mean(K)/h[i]
    f.bc  <- mean(M)/h[i]
    se.us <- sqrt((mean((K^2)) - mean(K)^2)/(N*h[i]^2))
    se.rb <- sqrt((mean((M^2)) - mean(M)^2)/(N*h[i]^2))

    if (kernel == "gau") {
      eN <- N
    } else {
      eN <- sum(abs(x - eval[i]) <= max(h[i], b[i]))
    }
    
    Estimate[i,] <- c(eval[i], h[i], b[i], eN, f.us, f.bc, se.us, se.rb) 
  }
  out<-list(Estimate=Estimate, opt=list(p=p, kernel=kernel.type, n=N, neval=neval, bwselect=bwselect))
  out$call <- match.call()
  class(out) <- "kdrobust"
  return(out)
}

print.kdrobust <- function(x,...){
  cat("Call: kdrobust\n\n")

  cat(paste("Sample size (n)                            =     ", x$opt$n,        "\n", sep=""))
  cat(paste("Kernel order for point estimation (p)      =     ", x$opt$p,        "\n", sep=""))
  cat(paste("Kernel function                            =     ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth method                           =     ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  invisible(x)
}

summary.kdrobust <- function(object, alpha = 0.05, sep = 5, ...) {
  z    <- qnorm(1 - alpha / 2)
  CI_l <- object$Estimate[, "tau.bc"] - object$Estimate[, "se.rb"] * z
  CI_r <- object$Estimate[, "tau.bc"] + object$Estimate[, "se.rb"] * z

  out <- list(opt      = object$opt,
              Estimate = object$Estimate,
              alpha    = alpha,
              sep      = sep,
              CI_l     = CI_l,
              CI_r     = CI_r)
  class(out) <- "summary.kdrobust"
  out
}

print.summary.kdrobust <- function(x, ...) {
  cat("Call: kdrobust\n\n")

  cat(paste("Sample size (n)                            =     ", x$opt$n,        "\n", sep=""))
  cat(paste("Kernel order for point estimation (p)      =     ", x$opt$p,        "\n", sep=""))
  cat(paste("Kernel function                            =     ", x$opt$kernel,   "\n", sep=""))
  cat(paste("Bandwidth selection method                 =     ", x$opt$bwselect, "\n", sep=""))
  cat("\n")

  alpha <- x$alpha
  sep   <- x$sep

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  cat(format(" ", width= 14 ))
  cat(format(" ", width= 10 ))
  cat(format(" ", width= 8  ))
  cat(format("Point", width= 10, justify="right"))
  cat(format("Std." , width= 10, justify="right"))
  cat(format("Robust B.C.", width=25, justify="centre"))
  cat("\n")

  cat(format("eval"            , width=14, justify="right"))
  cat(format("bw"              , width=10, justify="right"))
  cat(format("Eff.n"           , width=8 , justify="right"))
  cat(format("Est."            , width=10, justify="right"))
  cat(format("Error"           , width=10, justify="right"))
  cat(format(paste("[ ", floor((1-alpha)*100), "%", " C.I. ]", sep="")
             , width=25, justify="centre"))
  cat("\n")

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  for (j in 1:nrow(x$Estimate)) {
    cat(format(toString(j), width=4))
    cat(format(sprintf("%3.3f", x$Estimate[j, "eval"]), width=10, justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[j, "h"])  , width=10, justify="right"))
    cat(format(sprintf("%3.0f", x$Estimate[j, "N"])  , width=8 , justify="right"))
    cat(format(sprintf("%3.3f", x$Estimate[j, "tau.us"]) , width=10, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$Estimate[j, "se.us"]), sep=""), width=10, justify="right"))
    cat(format(paste("[", sprintf("%3.3f", x$CI_l[j]), " , ", sep="")  , width=14, justify="right"))
    cat(format(paste(sprintf("%3.3f", x$CI_r[j]), "]", sep=""), width=11, justify="left"))
    cat("\n")
    if (is.numeric(sep)) if (sep > 0) if (j %% sep == 0) {
      cat(paste(rep("-", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")
    }
  }

  cat(paste(rep("=", 14 + 10 + 8 + 10 + 10 + 25), collapse="")); cat("\n")

  invisible(x)
}

## kdrobust should approximate the true density for a large sample and
## integrate close to 1 over the support.

test_that("kdrobust recovers standard normal density at 0", {
  set.seed(1)
  x <- rnorm(30000)
  kd <- kdrobust(x, eval = 0, h = 0.3, kernel = "epa")
  expect_equal(as.numeric(kd$Estimate[, "tau.us"]), dnorm(0), tolerance = 0.01)
  expect_equal(as.numeric(kd$Estimate[, "tau.bc"]), dnorm(0), tolerance = 0.01)
})

test_that("kdrobust integrates to ~1 over a wide grid", {
  set.seed(1)
  x <- rnorm(20000)
  grid_step <- 0.1
  eval <- seq(-4, 4, by = grid_step)
  kd <- kdrobust(x, eval = eval, h = 0.4, kernel = "epa")
  integral <- sum(kd$Estimate[, "tau.us"]) * grid_step
  expect_equal(integral, 1, tolerance = 0.02)
})

test_that("kdrobust point estimate is non-negative for epa/uni", {
  set.seed(1)
  x <- rnorm(500)
  kd <- kdrobust(x, eval = seq(-2, 2, length.out = 20), h = 0.3, kernel = "epa")
  expect_true(all(kd$Estimate[, "tau.us"] >= 0))
})

test_that("kdrobust standard errors are positive and finite", {
  set.seed(1)
  x <- rnorm(500)
  kd <- kdrobust(x, eval = seq(-1, 1, length.out = 10))
  expect_true(all(is.finite(kd$Estimate[, "se.us"])))
  expect_true(all(kd$Estimate[, "se.us"] > 0))
  expect_true(all(is.finite(kd$Estimate[, "se.rb"])))
  expect_true(all(kd$Estimate[, "se.rb"] > 0))
})

\name{kdbwselect}
\alias{kdbwselect}
\title{Bandwidth Selection Procedures for Kernel Density Estimation and Inference}

\description{
\code{\link{kdbwselect}} implements bandwidth selectors for kernel density point estimators and inference procedures developed in Calonico, Cattaneo and Farrell (2018). See also Calonico, Cattaneo and Farrell (2022) for related optimality results.
It also implements other bandwidth selectors available in the literature.

Companion commands are: \code{\link{kdrobust}} for kernel density point estimation and inference procedures.

A detailed introduction to this command is given in Calonico, Cattaneo and Farrell (2019). For more details and related software useful for empirical analysis, visit \url{https://nppackages.github.io/}.
}
\usage{
kdbwselect(x, eval = NULL, neval = NULL, kernel = "epa",
bwselect = "mse-dpi", bwcheck=21, imsegrid=30, subset = NULL, data = NULL)
}
\arguments{
  \item{x}{independent variable.}
  
  \item{eval}{vector of evaluation point(s). By default it uses 30 quantile-spaced points (deciles 0.1 to 0.9 in equal steps) over the support of \code{x}.}

  \item{neval}{number of quantile-spaced evaluation points on the support of \code{x}. Default is \code{neval=30}.}
  
  \item{kernel}{kernel function used to construct the kernel estimators. Options are \code{epa} for the epanechnikov kernel, and \code{uni} for the uniform kernel. Default is \code{kernel = epa}.}
  
  \item{bwselect}{bandwidth selection procedure to be used. Options are:
  
  \code{mse-dpi} second-generation DPI implementation of MSE-optimal bandwidth. Default option.
	
	\code{imse-dpi} second-generation DPI implementation of IMSE-optimal bandwidth (computed using grid of evaluation points selected).
	
	\code{imse-rot} ROT implementation of IMSE-optimal bandwidth (computed using grid of evaluation points selected).
	
	\code{ce-dpi} second generation DPI implementation of CE-optimal bandwidth.
	
	\code{ce-rot} ROT implementation of CE-optimal bandwidth.
	
	\code{all} reports all available bandwidth selection procedures.
	
	Note: MSE = Mean Square Error; IMSE = Integrated Mean Squared Error; CE = Coverage Error; DPI = Direct Plug-in; ROT = Rule-of-Thumb. For details on implementation see Calonico, Cattaneo and Farrell (2019).}
	
  \item{bwcheck}{if a positive integer is provided, then the selected bandwidth is enlarged so that at least \code{bwcheck} effective observations are available at each evaluation point. Default is \code{bwcheck = 21}.}
  
  \item{imsegrid}{number of evaluations points used to compute the IMSE bandwidth selector. Default is \code{imsegrid = 30}.}
  
  \item{subset}{optional rule specifying a subset of observations to be used.}

  \item{data}{an optional data frame. When supplied, \code{x} and \code{subset} may be given as bare variable names referring to columns of \code{data}.}
}
\value{
\item{bws}{A matrix containing \code{eval} (grid points), \code{h} and \code{b} (selected bandwidths) for each evaluation point. When \code{bwselect="all"}, the matrix is widened to include both MSE and CER bandwidths from each of the DPI and ROT estimators.}
\item{bws.imse}{IMSE-optimal bandwidths returned when \code{bwselect="all"}; \code{NULL} otherwise.}
\item{opt}{A list containing options passed to the function.}
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2018. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf}{On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference}. Journal of the American Statistical Association, 113(522): 767-779. \doi{10.1080/01621459.2017.1285776}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2019. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf}{nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference}. Journal of Statistical Software, 91(8): 1-33. \doi{10.18637/jss.v091.i08}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2022. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2022_Bernoulli.pdf}{Coverage Error Optimal Confidence Intervals for Local Polynomial Regression}. Bernoulli, 28(4): 2998-3022.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.
}
\examples{
x   <- rnorm(500)
est <- kdbwselect(x)
summary(est)
}
\seealso{
  \code{\link{kdrobust}}
}

\name{nprobust-package}
\alias{nprobust-package}
\docType{package}
\alias{nprobust}
\title{Kernel Density and Local Polynomial Regression Methods}
\description{
   This package implements estimation, inference, bandwidth selection, and graphical procedures for kernel density and local polynomial regression methods, including robust bias-corrected confidence intervals as described in Calonico, Cattaneo and Farrell (2018): \code{\link{lprobust}} for local polynomial point estimation and robust bias-corrected inference, \code{\link{lpbwselect}} for local polynomial bandwidth selection, \code{\link{kdrobust}} for kernel density point estimation and robust bias-corrected inference, \code{\link{kdbwselect}} for kernel density bandwidth selection, and \code{\link{nprobust.plot}} for plotting results. The main methodological and numerical features are described in Calonico, Cattaneo and Farrell (2019).}
\details{
\tabular{ll}{
Package: \tab nprobust\cr
Type: \tab Package\cr
Version: \tab 1.0.0\cr
Date: \tab 2026-05-17\cr
License: \tab GPL-3\cr
}
Function for LPR estimation and inference: \code{\link{lprobust}}\cr
Function for LPR bandwidth selection: \code{\link{lpbwselect}}\cr
Function for KDE estimation and inference: \code{\link{kdrobust}}\cr
Function for KDE bandwidth selection: \code{\link{kdbwselect}}\cr
Function for graphical analysis: \code{\link{nprobust.plot}}\cr
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2018. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf}{On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference}. Journal of the American Statistical Association, 113(522): 767-779. \doi{10.1080/01621459.2017.1285776}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2019. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf}{nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference}. Journal of Statistical Software, 91(8): 1-33. \doi{10.18637/jss.v091.i08}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2022. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2022_Bernoulli.pdf}{Coverage Error Optimal Confidence Intervals for Local Polynomial Regression}. Bernoulli, 28(4): 2998-3022.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.
}

\name{lprobust}
\alias{lprobust}
\title{Local Polynomial Methods with Robust Bias-Corrected Inference}

\description{
\code{\link{lprobust}} implements local polynomial regression point estimators, with robust bias-corrected confidence intervals and inference procedures developed in Calonico, Cattaneo and Farrell (2018). See also Calonico, Cattaneo and Farrell (2022) for related optimality results.
It also implements other estimation and inference procedures available in the literature.

Companion commands: \code{\link{lpbwselect}} for local polynomial data-driven bandwidth selection, and \code{\link{nprobust.plot}} for plotting results.

A detailed introduction to this command is given in Calonico, Cattaneo and Farrell (2019). For more details and related software useful for empirical analysis, visit \url{https://nppackages.github.io/}.
}
\usage{
lprobust(y, x, eval = NULL, neval = NULL, p = NULL, deriv = NULL,
h = NULL, b = NULL, rho = 1, kernel = "epa", bwselect = NULL,
bwcheck = 21, bwregul = 1, imsegrid = 30, vce = "nn", covgrid = FALSE,
cluster = NULL, nnmatch = 3, level = 95, interior = FALSE, subset = NULL,
weights = NULL, masspoints = "check", data = NULL)
}
\arguments{
  \item{y}{dependent variable.}
  
  \item{x}{independent variable.}
  
  \item{eval}{vector of evaluation point(s). By default it uses 30 equally spaced points over the support of \code{x}.}

  \item{neval}{number of equally spaced evaluation points on the support of \code{x}. Default is \code{neval=30}.}
  
  \item{p}{polynomial order used to construct point estimator; default is \code{p = 1} (local linear regression).}
  
  \item{deriv}{derivative order of the regression function to be estimated. Default is \code{deriv=0} (regression function).}
  
  \item{h}{main bandwidth used to construct local polynomial point estimator. Can be either scalar (same bandwidth for all evaluation points), or vector of same dimension as \code{eval}. If not specified, bandwidth \code{h} is computed by the companion command \code{\link{lpbwselect}}.}
  
  \item{b}{bias bandwidth used to construct the bias-correction estimator. Can be either scalar (same bandwidth for all evaluation points), or vector of same dimension as \code{eval}. By default it is set equal to \code{h}. If \code{rho} is set to zero, \code{b} is computed by the companion command \code{\link{lpbwselect}}.}
  
  \item{rho}{Sets \code{b=h/rho}. Default is \code{rho = 1}.}
  
  \item{kernel}{kernel function used to construct local polynomial estimators. Options are \code{epa} for the epanechnikov kernel, \code{tri} for the triangular kernel, \code{uni} for the uniform kernel and \code{gau} for the gaussian kernel. Default is \code{kernel = epa}.}
  
  \item{bwselect}{bandwidth selection procedure to be used via \code{\link{lpbwselect}}. By default it computes \code{h} and sets \code{b=h/rho} (with \code{rho=1} by default). It computes both \code{h} and \code{b} if \code{rho} is set equal to zero. Options are:
  
  \code{mse-dpi} second-generation DPI implementation of MSE-optimal bandwidth. Default option if only one evaluation point is chosen.
	
	\code{mse-rot} ROT implementation of MSE-optimal bandwidth.
	
	\code{imse-dpi} second-generation DPI implementation of IMSE-optimal bandwidth (computed using a grid of evaluation points). Default option if more than one evaluation point is chosen.
	
	\code{imse-rot} ROT implementation of IMSE-optimal bandwidth (computed using a grid of evaluation points).
	
	\code{ce-dpi} second generation DPI implementation of CE-optimal bandwidth.
	
	\code{ce-rot} ROT implementation of CE-optimal bandwidth.

	Use \code{\link{lpbwselect}} with \code{bwselect="all"} to report all available bandwidth selection procedures.
	
	Note: MSE = Mean Square Error; IMSE = Integrated Mean Squared Error; CE = Coverage Error; DPI = Direct Plug-in; ROT = Rule-of-Thumb. For details on implementation see Calonico, Cattaneo and Farrell (2019).}
	
  \item{bwcheck}{if a positive integer is provided, then the selected bandwidth is enlarged so that at least \code{bwcheck} effective observations are available at each evaluation point. Default is \code{bwcheck = 21}.}
  
  \item{bwregul}{specifies scaling factor for the regularization term added to the denominator of bandwidth selectors. Setting \code{bwregul = 0} removes the regularization term from the bandwidth selectors. Default is \code{bwregul = 1}.}
  
  \item{imsegrid}{number of evaluations points used to compute the IMSE bandwidth selector. Default is \code{imsegrid = 30}.}
  
  \item{vce}{procedure used to compute the variance-covariance matrix estimator. Options are:

    \code{nn} heteroskedasticity-robust nearest neighbor variance estimator with \code{nnmatch} the (minimum) number of neighbors to be used. Default choice (when \code{cluster} is not supplied).

	\code{hc0} heteroskedasticity-robust plug-in residuals variance estimator without weights.

	\code{hc1} heteroskedasticity-robust plug-in residuals variance estimator with \code{hc1} weights.

	\code{hc2} heteroskedasticity-robust plug-in residuals variance estimator with \code{hc2} weights.

	\code{hc3} heteroskedasticity-robust plug-in residuals variance estimator with \code{hc3} weights.

	\code{cr1} cluster-robust variance estimator with \code{((n-1)/(n-k)) * (G/(G-1))} multiplier (Stata-style CR1). Default choice when \code{cluster} is supplied. Requires \code{cluster}.

	\code{cr2} cluster-robust variance estimator with Bell-McCaffrey block-adjusted residuals. Requires \code{cluster}.

	\code{cr3} cluster-robust variance estimator with block jackknife-style adjustment and \code{(G-1)/G} multiplier. Requires \code{cluster}.

	When \code{cluster} is supplied, \code{vce=hc0} or \code{vce=hc1} is remapped to \code{cr1}, \code{vce=hc2} to \code{cr2}, and \code{vce=hc3} to \code{cr3} (with a warning). \code{vce=nn} with \code{cluster} silently defaults to \code{cr1}.}

  \item{covgrid}{if TRUE, it computes two covariance matrices (cov.us and cov.rb) for classical and robust covariances across point estimators over the grid of evaluation points.}

  \item{cluster}{indicates the cluster ID variable used for cluster-robust variance estimation. When supplied, the default \code{vce} becomes \code{cr1}; \code{cr2} and \code{cr3} are also available.}

  \item{nnmatch}{to be combined with for \code{vce=nn} for heteroskedasticity-robust nearest neighbor variance estimator with \code{nnmatch} indicating the minimum number of neighbors to be used. Default is \code{nnmatch=3}}.
  
  \item{level}{confidence level used for confidence intervals; default is \code{level = 95}.}
  
  \item{interior}{if TRUE, all evaluation points are assumed to be interior points. This option affects only data-driven bandwidth selection via \code{\link{lpbwselect}}. Default is \code{interior = FALSE}.}
  
  \item{subset}{optional rule specifying a subset of observations to be used.}

  \item{weights}{optional vector of non-negative observation weights of the same length as \code{x}. User weights multiply the kernel weights in estimation, bandwidth selection, and variance computation (weighted least squares interpretation).}

  \item{masspoints}{how to handle evaluation points whose bandwidth window contains few unique \code{x} values (the local polynomial may become unreliable in that case). Options: \code{"check"} (default) warns when there are fewer than \code{p+5} unique values inside the main bandwidth; \code{"off"} disables the check.}

  \item{data}{an optional data frame. When supplied, \code{y}, \code{x}, \code{cluster}, \code{weights}, and \code{subset} may be given as bare variable names referring to columns of \code{data}.}
}
\value{
\item{Estimate}{A matrix containing \code{eval} (grid points), \code{h}, \code{b} (bandwidths), \code{N}
  (effective sample sizes), \code{tau.us} (point estimates with p-th order local polynomial),
  \code{tau.bc} (bias-corrected point estimates with (p+1)-th order local polynomial),
  \code{se.us} (standard error corresponding to \code{tau.us}), and \code{se.rb} (robust standard error).}
\item{cov.us}{Conventional-estimator covariance matrix across the evaluation grid (\code{neval} by \code{neval}). Returned only when \code{covgrid=TRUE}; \code{NULL} otherwise.}
\item{cov.rb}{Robust-bias-corrected covariance matrix across the evaluation grid (\code{neval} by \code{neval}). Returned only when \code{covgrid=TRUE}; \code{NULL} otherwise.}
\item{opt}{A list containing options passed to the function.}
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2018. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf}{On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference}. Journal of the American Statistical Association, 113(522): 767-779. \doi{10.1080/01621459.2017.1285776}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2019. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf}{nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference}. Journal of Statistical Software, 91(8): 1-33. \doi{10.18637/jss.v091.i08}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2022. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2022_Bernoulli.pdf}{Coverage Error Optimal Confidence Intervals for Local Polynomial Regression}. Bernoulli, 28(4): 2998-3022.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.
}
\examples{
x   <- runif(500)
y   <- sin(4*x) + rnorm(500)
est <- lprobust(y,x)
summary(est)
}
\keyword{ LPR }
\keyword{ Robust Estimation }
\seealso{
  \code{\link{lpbwselect}}
}

\name{nprobust.plot}
\alias{nprobust.plot}
\title{Graphical Presentation of Results from \code{nprobust} Package.}
\description{
\code{nprobust.plot} plots estimated density and regression function using the \code{nprobust} package. A detailed introduction to this command is given in Calonico, Cattaneo and Farrell (2019).

Companion commands: \code{\link{lprobust}} for local polynomial point estimation and inference procedures, and \code{\link{kdrobust}} for kernel density point estimation and inference procedures.

For more details and related software useful for empirical analysis, visit \url{https://nppackages.github.io/}.
}
\usage{
nprobust.plot(..., alpha = NULL, type = NULL, CItype = NULL,
  title = "", xlabel = "", ylabel = "", lty = NULL, lwd = NULL,
  lcol = NULL, pty = NULL, pwd = NULL, pcol = NULL, CIshade = NULL,
  CIcol = NULL, legendTitle = NULL, legendGroups = NULL)
}
\arguments{
\item{...}{Objects returned by \code{\link{kdrobust}} or \code{\link{lprobust}}.}

\item{alpha}{Numeric scalar between 0 and 1, the significance level for plotting
confidence regions. If more than one is provided, they will be applied to data series
accordingly.}

\item{type}{String, one of \code{"line"} (default), \code{"points"} or \code{"both"}, how
the point estimates are plotted. If more than one is provided, they will be applied to data series
accordingly.}

\item{CItype}{String, one of \code{"region"} (shaded region, default), \code{"line"} (dashed lines),
\code{"ebar"} (error bars), \code{"all"} (all of the previous) or \code{"none"} (no confidence region),
how the confidence region should be plotted. If more than one is provided, they will be applied to data series
accordingly.}

\item{title, xlabel, ylabel}{Strings, title of the plot and labels for x- and y-axis.}

\item{lty}{Line type for point estimates, only effective if \code{type} is \code{"line"} or
\code{"both"}. \code{1} for solid line, \code{2} for dashed line, \code{3} for dotted line.
For other options, see the instructions for \code{\link[ggplot2]{ggplot2}} or \code{\link{par}}. If
more than one is provided, they will be applied to data series accordingly.}

\item{lwd}{Line width for point estimates, only effective if \code{type} is \code{"line"} or
\code{"both"}. Should be strictly positive. For other options, see the instructions for
\code{\link[ggplot2]{ggplot2}} or \code{\link{par}}. If more than one is provided, they will be applied
to data series accordingly.}

\item{lcol}{Line color for point estimates, only effective if \code{type} is \code{"line"} or
\code{"both"}. \code{1} for black, \code{2} for red, \code{3} for green, \code{4} for blue.
For other options, see the instructions for \code{\link[ggplot2]{ggplot2}}or \code{\link{par}}. If
more than one is provided, they will be applied to data series
accordingly.}

\item{pty}{Scatter plot type for point estimates, only effective if \code{type} is \code{"points"} or
\code{"both"}. For options, see the instructions for \code{\link[ggplot2]{ggplot2}} or \code{\link{par}}. If
more than one is provided, they will be applied to data series
accordingly.}

\item{pwd}{Scatter plot size for point estimates, only effective if \code{type} is \code{"points"} or
\code{"both"}. Should be strictly positive. If more than one is provided, they will be applied to data series
accordingly.}

\item{pcol}{Scatter plot color for point estimates, only effective if \code{type} is \code{"points"} or
\code{"both"}. \code{1} for black, \code{2} for red, \code{3}
for green, \code{4} for blue.
For other options, see the instructions for \code{\link[ggplot2]{ggplot2}} or \code{\link{par}}. If
more than one is provided, they will be applied to data series
accordingly.}

\item{CIshade}{Numeric, opaqueness of the confidence region, should be between 0 (transparent) and
1. Default is 0.2. If more than one is provided, they will be applied to data series
accordingly.}

\item{CIcol}{color for confidence region. \code{1} for black, \code{2} for red, \code{3}
for green, \code{4} for blue.
For other options, see the instructions for \code{\link[ggplot2]{ggplot2}} or \code{\link{par}}. If
more than one is provided, they will be applied to data series
accordingly.}
\item{legendTitle}{String, title of legend.}
\item{legendGroups}{String Vector, group names used in legend.}
}
\value{
A standard \code{\link[ggplot2]{ggplot2}} object is returned, hence can be used for further customization.
}
\details{
Companion commands: \code{\link{lprobust}} for local polynomial point estimation and inference, and \code{\link{kdrobust}} for kernel density point estimation and inference.
}
\examples{
x   <- runif(500) 
y   <- sin(4*x) + rnorm(500)
est <- lprobust(y,x)
nprobust.plot(est)
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2019. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf}{nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference}. Journal of Statistical Software, 91(8): 1-33. \doi{10.18637/jss.v091.i08}.
}
\seealso{
\code{\link{lprobust}}, \code{\link{kdrobust}}, \code{\link[ggplot2]{ggplot2}}
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.
}

\name{kdrobust}
\alias{kdrobust}
\title{Kernel Density Methods with Robust Bias-Corrected Inference}

\description{
\code{\link{kdrobust}} implements kernel density point estimators, with robust bias-corrected confidence intervals and inference procedures developed in Calonico, Cattaneo and Farrell (2018). See also Calonico, Cattaneo and Farrell (2022) for related optimality results.
It also implements other estimation and inference procedures available in the literature.

Companion commands: \code{\link{kdbwselect}} for kernel density data-driven bandwidth selection, and \code{\link{nprobust.plot}} for plotting results.

A detailed introduction to this command is given in Calonico, Cattaneo and Farrell (2019). For more details and related software useful for empirical analysis, visit \url{https://nppackages.github.io/}.
}
\usage{
kdrobust(x, eval = NULL, neval = NULL, h = NULL, b = NULL, rho = 1, 
kernel = "epa", bwselect = NULL, bwcheck = 21, imsegrid=30, level = 95, subset = NULL,
data = NULL)
}
\arguments{
  \item{x}{independent variable.}
  
  \item{eval}{vector of evaluation point(s). By default it uses 30 quantile-spaced points (deciles 0.1 to 0.9 in equal steps) over the support of \code{x}.}

  \item{neval}{number of quantile-spaced evaluation points on the support of \code{x}. Default is \code{neval=30}.}
  
  \item{h}{main bandwidth used to construct the kernel density point estimator. Can be either scalar (same bandwidth for all evaluation points), or vector of same dimension as \code{eval}. If not specified, bandwidth \code{h} is computed by the companion command \code{\link{kdbwselect}}.}
  
  \item{b}{bias bandwidth used to construct the bias-correction estimator. Can be either scalar (same bandwidth for all evaluation points), or vector of same dimension as \code{eval}. By default it is set equal to \code{h}. If \code{rho} is set to zero, \code{b} is computed by the companion command \code{\link{kdbwselect}}.}
  
  \item{rho}{Sets \code{b=h/rho}. Default is \code{rho = 1}.}
  
  \item{kernel}{kernel function used to construct kernel density estimators. Options are \code{epa} for the epanechnikov kernel and \code{uni} for the uniform kernel. Default is \code{kernel = epa}.}
  
  \item{bwselect}{bandwidth selection procedure to be used via \code{\link{kdbwselect}}. By default it computes \code{h} and sets \code{b=h/rho} (with \code{rho=1} by default). It computes both \code{h} and \code{b} if \code{rho} is set equal to zero. Options are:
  
  \code{mse-dpi} second-generation DPI implementation of MSE-optimal bandwidth. Default option if only one evaluation point is chosen.
	
	\code{imse-dpi} second-generation DPI implementation of IMSE-optimal bandwidth (computed using a grid of evaluation points). Default option if more than one evaluation point is chosen.
	
	\code{imse-rot} ROT implementation of IMSE-optimal bandwidth (computed using a grid of evaluation points).
	
	\code{ce-dpi} second generation DPI implementation of CE-optimal bandwidth.
	
	\code{ce-rot} ROT implementation of CE-optimal bandwidth.

	Use \code{\link{kdbwselect}} with \code{bwselect="all"} to report all available bandwidth selection procedures.
	
	Note: MSE = Mean Square Error; IMSE = Integrated Mean Squared Error; CE = Coverage Error; DPI = Direct Plug-in; ROT = Rule-of-Thumb. For details on implementation see Calonico, Cattaneo and Farrell (2019).}
	
  \item{bwcheck}{if a positive integer is provided, then the selected bandwidth is enlarged so that at least \code{bwcheck} effective observations are available at each evaluation point. Default is \code{bwcheck = 21}.}
  
  \item{imsegrid}{number of evaluations points used to compute the IMSE bandwidth selector. Default is \code{imsegrid = 30}.}
  
  \item{level}{confidence level used for confidence intervals; default is \code{level = 95}.}
  
  \item{subset}{optional rule specifying a subset of observations to be used.}

  \item{data}{an optional data frame. When supplied, \code{x} and \code{subset} may be given as bare variable names referring to columns of \code{data}.}
}
\value{
\item{Estimate}{A matrix containing \code{eval} (grid points), \code{h}, \code{b} (bandwidths), \code{N}
  (effective sample sizes), \code{tau.us} (conventional density estimate),
  \code{tau.bc} (bias-corrected density estimate),
  \code{se.us} (standard error corresponding to \code{tau.us}), and \code{se.rb} (robust standard error).}
\item{opt}{A list containing options passed to the function.}
}
\references{
Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2018. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2018_JASA.pdf}{On the Effect of Bias Estimation on Coverage Accuracy in Nonparametric Inference}. Journal of the American Statistical Association, 113(522): 767-779. \doi{10.1080/01621459.2017.1285776}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2019. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2019_JSS.pdf}{nprobust: Nonparametric Kernel-Based Estimation and Robust Bias-Corrected Inference}. Journal of Statistical Software, 91(8): 1-33. \doi{10.18637/jss.v091.i08}.

Calonico, S., M. D. Cattaneo, and M. H. Farrell. 2022. \href{https://nppackages.github.io/references/Calonico-Cattaneo-Farrell_2022_Bernoulli.pdf}{Coverage Error Optimal Confidence Intervals for Local Polynomial Regression}. Bernoulli, 28(4): 2998-3022.
}
\author{
Sebastian Calonico, University of California, Davis, CA.  \email{scalonico@ucdavis.edu}.

Matias D. Cattaneo, Princeton University, Princeton, NJ.  \email{matias.d.cattaneo@gmail.com}.

Max H. Farrell, University of California, Santa Barbara, CA. \email{mhfarrell@gmail.com}.
}
\examples{
x   <- rnorm(500)
est <- kdrobust(x)
summary(est)
}
\keyword{ LPR }
\keyword{ Robust Estimation }
\seealso{
  \code{\link{kdbwselect}}
}
